# encoding: utf-8
#
#Copyright (C) 2017-2023, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
python implementations of the Green's Dyads, accelerated using `numba`
"""
from __future__ import print_function
from __future__ import absolute_import

import copy
import warnings

import numpy as np
import math
import cmath

import numba
# from numba import cuda

## bessel functions to calc. hankel functions (required for 2D green tensor)
# import numba_scipy
from scipy.special import jv, yv
from scipy.special import iv, kv

from pyGDM2.propagators import propagators



#==============================================================================
# globals
#==============================================================================
DTYPE_C = np.complex64


# =============================================================================
#               numba overloads for Bessel functions
#
# adapted for scipy>=1.4 from https://github.com/numba/numba-scipy
#
# =============================================================================
import ctypes
import scipy.special
from numba.extending import get_cython_function_address


## -- in future: add binding to complex hankel
## double complex hankel1(double, double complex)

name_to_numba_signatures = {
    'yv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    'jv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    
    'iv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    'kv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    }

name_and_types_to_pointer = {
    ('yv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1yv')),
    ('jv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1jv')),
    
    ('iv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1iv')),
    ('kv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1kv')),
    }

def choose_kernel(name, all_signatures):
    def choice_function(*args):
        for signature in all_signatures:
            if args == signature:
                f = name_and_types_to_pointer[(name, *signature)]
                return lambda *args: f(*args)
    return choice_function

def add_overloads():
    for name, all_signatures in name_to_numba_signatures.items():
        sc_function = getattr(scipy.special, name)
        numba.extending.overload(sc_function)(
            choose_kernel(name, all_signatures)
        )

add_overloads()



### --- non retarded 2D
# =============================================================================
# propagators for infinitely long dipole line, 2D
# 
# The 2D Green's tensors implementation is based on fortran code 
# written by G. Colas des Francs
#   
# requires numba-scipy (pip install numba-scipy)
# =============================================================================
@numba.njit()
def _numba_hankel1(n, x):
    ## numba-compatible first kind hankel functions
    ## only scipy-bessel functions are numba-compatible (via `numba-scipy` package)
    ## numba binding supports only real arguments so far (as of March 2020)
    
    ## purely real argument: Use Bessel functions
    if np.imag(x) == 0:
        Jn = jv(n, np.real(x))
        Yn = yv(n, np.real(x))
    
    ## purely imaginary argument: Use modified Bessel functions (first kind: I, second kind: K)
    elif np.real(x) == 0:
        ## identities: (modif. Bessel first kind: I, second kind: K)
        ## Jn(1j*z) = exp(n*1j*np.pi/2) * In(z)
        ## Jn(1j*z) = (j^n) * In(z)
        ## Yn(1j*z) = exp((n+1)*1j*np.pi/2) * In(z) - 2/np.pi * exp(-1*n*1j*np.pi/2) * Kn(z)
        Jn = (1j**n) * iv(n, np.imag(x))
        Yn = cmath.exp((n+1)*1j*np.pi/2) * iv(n, np.imag(x)) - \
              2/np.pi * cmath.exp(-1*n*1j*np.pi/2) * kv(n, np.imag(x))
        
    ## 'mixed' complex argument: not possible with numba
    else:
        raise Exception("2D Greens tensor calculation: " +
                        "Numba-hankel functions support only purely real / purely imaginary argument.")
    
    h1n = Jn + 1j*Yn
    return h1n


@numba.njit()
def _s0_2D(Dx, Dz, wavelength, eps, ky0=0):
    ## the actual free space dyad calculation via dipole<-->observation distance
    eps = np.real(eps)   # numba bessel support only real arguments
    
    k0 = 2*np.pi / wavelength
    k02 = k0**2
    ## only real epsilon supported!
    kr = cmath.sqrt(eps * k02 - ky0**2)
    
    lR = math.sqrt(Dx**2 + Dz**2)
    if lR < 0.01:
       lR = 0.0
       raise ValueError("Points too close. Diverging free-space 2D Green's Tensor!")
    
    cop = Dx / lR
    sip = Dz / lR
    co2p = 2. * cop**2 - 1.0
    si2p = 2. * cop*sip
    
    h01 = _numba_hankel1(0.0, kr*lR)
    h11 = _numba_hankel1(1.0, kr*lR)
    h21 = _numba_hankel1(2.0, kr*lR)
    
    gxx = ( (k02 - kr**2 / (2.0*eps)) * h01 +
             kr**2 * co2p * h21 / (2.0*eps) )
    gxy = -1j * kr * ky0 * cop * h11 / eps
    gxz = (kr**2 * si2p)*h21 / (2.0*eps)  
    
    gyx = gxy
    gyy = (k02 - ky0**2/eps) * h01   
    gyz = -1j * kr * ky0 * sip * h11 / eps     
    
    gzx = gxz
    gzy = gyz
    gzz = ( (k02 - kr**2 / (2.0*eps)) * h01 -
             kr**2 * co2p * h21 / (2.0*eps) )
    
    ## cgs!
    return 1j*np.pi*gxx, 1j*np.pi*gyy, 1j*np.pi*gzz, \
           1j*np.pi*gxy, 1j*np.pi*gxz, 1j*np.pi*gyx, \
           1j*np.pi*gyz, 1j*np.pi*gzx, 1j*np.pi*gzy
    ## SI: *1/(4 pi)


    
@numba.njit()      
def _G0_2D(R1, R2, wavelength, eps, ky0=0):
    ## ky0=0.0: so far incidence || XZ plane
    
    Dx = R2[0] - R1[0]
    Dz = R2[2] - R1[2]
    # lR = math.sqrt(Dx**2 + Dz**2)
    
    return _s0_2D(Dx, Dz, wavelength, eps, ky0)


## 1-2-3 surface dyad via image-charge approximation
@numba.njit()
def _G_mirrorcharge_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=0.0):
    Dx = R2[0] - R1[0]
    
    ## init propagator variables
    sxx123, syy123, szz123, sxy123, sxz123, \
           syx123, syz123, szx123, szy123 = np.zeros(9, dtype=np.complex64)
    
    ## interface 1-2, evaluate only if present
    if eps1 != eps2:
        ## mirror charge z position 
        Dz12 = R2[2] + R1[2]
        
        if R1[2] <= 0:
            eps_dipole1 = eps2
            cdelta12 = (eps2-eps1)/(eps1+eps2)
        elif spacing >= R1[2] > 0:
            eps_dipole1 = eps1
            cdelta12 = (eps1-eps2)/(eps1+eps2)
        elif R1[2] > spacing:
            eps_dipole1 = eps1
            cdelta12 = (eps1-eps2)/(eps1+eps2)
        
        ## propagator
        sxx, syy, szz, sxy, sxz, \
              syx, syz, szx, szy = _s0_2D(Dx, Dz12, wavelength, eps_dipole1, ky0)
        sxx123 += -1*cdelta12 * sxx
        syy123 += -1*cdelta12 * syy
        szz123 += +1*cdelta12 * szz
        sxy123 += -1*cdelta12 * sxy
        sxz123 += -1*cdelta12 * sxz
        syx123 += -1*cdelta12 * syx
        syz123 += -1*cdelta12 * syz
        szx123 += -1*cdelta12 * szx
        szy123 += -1*cdelta12 * szy
        
    ## interface 2-3, evaluate only if present
    if eps2 != eps3:
        ## mirror charge z position 
        Dz23 = R2[2] + R1[2] - 2*spacing
        
        if R1[2] <= 0:
            eps_dipole2 = eps3
            cdelta23 = (eps3-eps2)/(eps3+eps2)
        elif spacing >= R1[2] > 0:
            eps_dipole2 = eps3
            cdelta23 = (eps3-eps2)/(eps3+eps2)
        elif R1[2] > spacing:
            eps_dipole2 = eps2
            cdelta23 = (eps2-eps3)/(eps3+eps2)
            
        ## propagator
        sxx, syy, szz, sxy, sxz, \
              syx, syz, szx, szy = _s0_2D(Dx, Dz23, wavelength, eps_dipole2, ky0)
        sxx123 += -1*cdelta23 * sxx
        syy123 += -1*cdelta23 * syy
        szz123 += +1*cdelta23 * szz
        sxy123 += -1*cdelta23 * sxy
        sxz123 += -1*cdelta23 * sxz
        syx123 += -1*cdelta23 * syx
        syz123 += -1*cdelta23 * syz
        szx123 += -1*cdelta23 * szx
        szy123 += -1*cdelta23 * szy
        
    return sxx123, syy123, szz123, \
           sxy123, sxz123, syx123, \
           syz123, szx123, szy123
    

@numba.njit()
def _G0_HE_2D(R1, R2, wavelength, eps, ky0=0):
    ## the actual free space dyad calculation via dipole<-->observation distance
    eps = np.real(eps)   # numba bessel support only real arguments
    Dx = R2[0] - R1[0]
    Dz = R2[2] - R1[2]
    
    k0 = 2*np.pi / wavelength
    k02 = k0**2
    ## only real epsilon supported!
    kr = cmath.sqrt(eps * k02 - ky0**2)
    
    lR = math.sqrt(Dx**2 + Dz**2)
    if lR < 0.01:
       lR = 0.0
       raise ValueError("Points too close. Diverging free-space 2D Green's Tensor!")
    
    if ky0 != 0:
        h01 = _numba_hankel1(0.0, kr*lR)
    else:
        h01 = 0
    h11 = _numba_hankel1(1.0, kr*lR)
    
    phase_y = np.exp(1j * ky0 * R2[1])
    
    gxy = k0 * np.pi * h11 * (kr / lR) * Dz * phase_y
    gxz = 1j * np.pi * k0 * ky0 * h01 * phase_y
    gyz = k0 * np.pi * h11 * (kr / lR) * Dx * phase_y
    
    gyx = -gxy
    gzx = -gxz
    gzy = -gyz
    
    gxx = 0
    gyy = 0
    gzz = 0
    
    return gxx, gyy, gzz, \
           gxy, gxz, gyx, \
           gyz, gzx, gzy



## --- the full propagator: vacuum + surface term -- nearfield approximation!
##     assuming emitter in center layer, to calcualte resulting field anywhere
@numba.njit()
def G0_EE_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=0):
    """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
    R1: dipole position
    R2: evaluation position
    """
    if eps1==eps2 and eps2==eps3:
        eps = eps2
        scr_f = 1.0
    else:
        ## ----- emitter dipole situated in substrate layer ('eps1')
        if R1[2] < 0:
            ## field evaluated in bottom layer ("substrate")
            if R2[2] < 0:
                eps = eps1
                scr_f = 1.0
            ## field evaluated in center layer (same as emitter dipole)
            elif 0 <= R2[2] <= spacing:
                eps = eps2
                scr_f = eps * 2.0/(eps1+eps2)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in top cladding layer
            elif R2[2] > spacing:
                eps = eps3
                scr_f = eps * 2.0/(eps1+eps3)   # factor 'eps' to take 1/eps in G0 into account
        
        ## ----- emitter dipole situated in center layer ('eps2')
        if 0 <= R1[2] <= spacing:
            ## field evaluated in bottom layer ("substrate")
            if R2[2] < 0:
                eps = eps1
                scr_f = eps * 2.0/(eps1+eps2)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in center layer (same as emitter dipole)
            elif 0 <= R2[2] <= spacing:
                eps = eps2
                scr_f = 1.0
            ## field evaluated in top cladding layer
            elif R2[2] > spacing:
                eps = eps3
                scr_f = eps * 2.0/(eps2+eps3)   # factor 'eps' to take 1/eps in G0 into account
        
        ## ----- emitter dipole situated in top cladding layer ('eps3')
        if R1[2] > spacing:
            ## field evaluated in bottom layer ("substrate")
            if R2[2] < 0:
                eps = eps1
                scr_f = eps * 2.0/(eps1+eps3)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in center layer (same as emitter dipole)
            elif 0 <= R2[2] <= spacing:
                eps = eps2
                scr_f = eps * 2.0/(eps2+eps3)   # factor 'eps' to take 1/eps in G0 into account
            ## field evaluated in top cladding layer
            elif R2[2] > spacing:
                eps = eps3
                scr_f = 1.0
    
    ## ----- free space dyad
    xx, yy, zz, xy, xz, yx, yz, zx, zy = _G0_2D(R1, R2, wavelength, eps, ky0)

    return scr_f*xx, scr_f*yy, scr_f*zz, \
           scr_f*xy, scr_f*xz, scr_f*yx, \
           scr_f*yz, scr_f*zx, scr_f*zy


@numba.njit()
def Gs_EE_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=0):
    """evaluate field of mirror-charge in 1-2-3 layer environment (NF approx.)
    
    R1: dipole position
    R2: evaluation position
    """
    xx, yy, zz, xy, xz, yx, yz, zx, zy = [np.complex64(0)]*9
    
    ## ----- actual interface present
    if eps1!=eps2 or eps2!=eps3:
        ## mirror charges NF Green's tensor
        xx, yy, zz, xy, xz, yx, yz, zx, zy = _G_mirrorcharge_123_2D(R1, R2, wavelength, 
                                                    eps1, eps2, eps3, spacing, ky0)

    return xx, yy, zz, xy, xz, yx, yz, zx, zy


@numba.njit()
def Gtot_EE_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=0):
    """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
    R1: dipole position
    R2: evaluation position
    """
    ## ----- free space term
    xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_EE_123_2D(R1, R2, wavelength, 
                                                   eps1, eps2, eps3, spacing, ky0)
    
    ## ----- surface term
    xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_EE_123_2D(R1, R2, wavelength, 
                                                    eps1, eps2, eps3, spacing, ky0)

    return xx + xxs, yy + yys, zz + zzs, \
           xy + xys, xz + xzs, yx + yxs, \
           yz + yzs, zx + zxs, zy + zys

# @numba.njit()
# def G_EE_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=0.0):
#     """TODO: G0+Gs, conditions for source / observation position"""
#     ## ky0=0.0: so far incidence || XZ plane
#     xx, yy, zz, xy, xz, yx, yz, zx, zy = _G0_2D(R1, R2, wavelength, eps2)
    
#     xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = _Gs123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing)
    
#     ## dummy screening factor
#     scr_f = 1.0
#     return scr_f*xx + xxs, scr_f*yy + yys, scr_f*zz + zzs, \
#            scr_f*xy + xys, scr_f*xz + xzs, scr_f*yx + yxs, \
#            scr_f*yz + yzs, scr_f*zx + zxs, scr_f*zy + zys



@numba.njit()
def G_HE_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=0.0):
    """TODO: G0+Gs, conditions for source / observation position"""
    ## ky0=0.0: so far incidence || XZ plane. Should be added to parameters
    
    ## field evaluated in bottom layer ("substrate")
    if R2[2] < 0:
        eps = eps1
    ## field evaluated in center layer (same as emitter dipole)
    elif 0 <= R2[2] <= spacing:
        eps = eps2
    ## field evaluated in top cladding layer
    elif spacing <= R2[2]:
        eps = eps3
        
    xx, yy, zz, xy, xz, yx, yz, zx, zy = _G0_HE_2D(R1, R2, wavelength, eps, ky0)
    
    ## dummy screening factor
    return xx, yy, zz, \
           xy, xz, yx, \
           yz, zx, zy


# =============================================================================
# ensemble-propagator routines
# =============================================================================
## --- multi-dipole field propagation
@numba.njit(parallel=True)
def greens_tensor_evaluation_2D(dp_pos, r_probe, G_func, wavelength, conf_dict, M,
                                selfterm=np.zeros((3,3)).astype(np.complex64), 
                                dist_div_G=0.5):
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    ky0 = np.float32(conf_dict['ky0'].real)
    
    for i in numba.prange(len(dp_pos)):   # explicit parallel loop
        _pos = dp_pos[i]
        for j in range(len(r_probe)):
            _r = r_probe[j]
            if np.sqrt((_r[0]-_pos[0])**2 + (_r[1]-_pos[1])**2 + (_r[2]-_pos[2])**2)<dist_div_G:
                xx, xy, xz = selfterm[0]
                yx, yy, yz = selfterm[1]
                zx, zy, zz = selfterm[2]
            else:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = G_func(_pos, _r, wavelength, 
                                                            eps1, eps2, eps3, spacing, ky0)
            ## return list of Greens tensors
            M[i,j,0,0], M[i,j,1,1], M[i,j,2,2] = xx, yy, zz
            M[i,j,1,0], M[i,j,2,0], M[i,j,0,1] = yx, zx, xy
            M[i,j,2,1], M[i,j,0,2], M[i,j,1,2] = zy, xz, yz




### --- coupling matrix setup
@numba.njit(parallel=True)
def t_sbs_EE_123_2D(geo, wavelength, selfterms, alpha, conf_dict, M):
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    ky0 = np.float32(conf_dict['ky0'].real)    
    
# =============================================================================
#     SBS coupling-matrix setup with 2d propagators
# =============================================================================
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]
        for j in range(len(geo)):
            R1 = geo[j]
            aj = alpha[j]
            st = selfterms[j]
            ## --- vacuum dyad
            if i==j:
                ## self term
                xx, yy, zz = st[0,0], st[1,1], st[2,2]
                xy, xz, yx = st[0,1], st[0,2], st[1,0]
                yz, zx, zy = st[1,2], st[2,0], st[2,1]
            else:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_EE_123_2D(R1, R2, wavelength, 
                                                       eps1, eps2, eps3, spacing, ky0)
            
            ## --- 1-2-3 surface dyad (non retarded NF approximation)
            if eps1!=eps2 or eps2!=eps3:
                xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_EE_123_2D(
                            R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0)
                ## combined dyad
                xx, yy, zz, xy, xz, yx, yz, zx, zy = xx+xxs, yy+yys, zz+zzs, \
                                                     xy+xys, xz+xzs, yx+yxs, \
                                                     yz+yzs, zx+zxs, zy+zys
            
            ## return invertible matrix:  delta_ij*1 - G[i,j] * alpha[j]
            M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
            M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
            M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
            M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
            M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
            M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
            M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
            M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
            M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])
            if i==j:
                M[3*i+0, 3*j+0] += 1
                M[3*i+1, 3*j+1] += 1
                M[3*i+2, 3*j+2] += 1

@numba.njit(parallel=True)
def t_sbs_HE_123_2D(geo, wavelength, selfterms, alpha, conf_dict, M):
    eps2 = conf_dict['eps2']
    ky0 = np.float32(conf_dict['ky0'].real)
    
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]        # "observer"
        for j in numba.prange(len(geo)):
            R1 = geo[j]    # emitter
            aj = alpha[j]
            st = selfterms[j]
            ## --- vacuum dyad
            if i==j:
                ## self term
                xx, yy, zz = st[0,0], st[1,1], st[2,2]
                xy, xz, yx = st[0,1], st[0,2], st[1,0]
                yz, zx, zy = st[1,2], st[2,0], st[2,1]
            else:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = _G0_HE_2D(R1, R2, wavelength, eps2, ky0)
            
            ## return: G[i,j] * alpha[j]
            ## --- magnetic-electric part
            M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
            M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
            M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
            M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
            M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
            M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
            M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
            M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
            M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])




### --- dyad class 2D non-retarded
class DyadsQuasistatic2D123(propagators.DyadsBaseClass):
    __name__ = "3-layer environment: Quasistatic 2D '1-2-3' Green's tensors"
    
    def __init__(self, n1=None, n2=None, n3=None, spacing=5000, ky0=None,
                 ignore_incident_field_error=False):
        super().__init__()
        
        ## Dyads
        self.G_EE = Gtot_EE_123_2D   # !!! todo: test the 1-2-3 interface tensor
        self.G_HE = G_HE_123_2D
        self.G_EE_ff = Gtot_EE_123_2D   # !!! todo: 2d-ff propagator here
        
        ## evaluate propagator routine
        self.eval_G = greens_tensor_evaluation_2D
        
        ## coupling matrix constructor routines
        self.tsbs_EE = t_sbs_EE_123_2D
        self.tsbs_HE = t_sbs_HE_123_2D
        
        from pyGDM2 import materials
        if isinstance(n1, (int, float, complex)) and not isinstance(n1, bool):
            self.n1_material = materials.dummy(n1)
        else:
            self.n1_material = n1
            
        n2 = n2 or n1     # if None, use `n1`
        if isinstance(n2, (int, float, complex)) and not isinstance(n2, bool):
            from pyGDM2 import materials
            self.n2_material = materials.dummy(n2)
        else:
            self.n2_material = n2
            
        n3 = n3 or n2     # if None, use `n2`
        if isinstance(n3, (int, float, complex)) and not isinstance(n3, bool):
            from pyGDM2 import materials
            self.n3_material = materials.dummy(n3)
        else:
            self.n3_material = n3
        
        self.spacing = spacing
        self.ky0 = ky0
        self.ignore_incident_field_error = ignore_incident_field_error
        
    def __repr__(self, verbose=False):
        """description about simulation environment defined by set of dyads"""
        out_str =  ' ------ environment -------'
        out_str += '\n ' + self.__name__
        out_str += '\n '
        out_str += '\n' + '   n3 = {}  <-- top'.format(self.n3_material.__name__)
        out_str += '\n' + '   n2 = {}  <-- structure zone (height "spacing" = {}nm)'.format(
                        self.n2_material.__name__, self.spacing)
        out_str += '\n' + '   n1 = {}  <-- substrate'.format(self.n1_material.__name__)
        return out_str
    
    
    def exceptionHandling(self, struct, efield):
        """Exception handling / consistency check for the set of tensors
        
        check if structure and incident field generator are compatible

        Parameters
        ----------
        struct : :class:`.structures.struct`
            instance of structure class
        efield : :class:`.fields.efield`
            instance of incident field class

        Returns
        -------
        bool : True if struct and field are compatible, False if they don't fit the tensors

        """
        from pyGDM2 import fields_py
        if efield.field_generator not in [fields_py.planewave, 
                                          fields_py.evanescent_planewave,
                                          fields_py.plane_wave,
                                          field_dipole_line_2d]:
            if self.ignore_incident_field_error:
                warnings.warn('A non-2D incident field is used! This may lead to incorrect results.')
            else:
                raise Exception("Incompatible incident field. "+
                                "2D Dyads are only compatible with fields `plane_wave` and `planewave`."+
                                "Ignore this error voia the `ignore_incident_field_error` argument.")
        
        all_inc_angles = []
        for k in efield.kwargs_permutations:
            if 'inc_plane' in k.keys():
                if k['inc_plane'].lower() in ['yz', 'zy'] and 'theta_inc' in k.keys():
                    all_inc_angles.append(k['theta_inc'])
            if 'ky0' in k.keys():
                all_inc_angles.append(k['ky0'])
        if len(np.unique(all_inc_angles)) > 1:
            raise Exception("2D simulations can treat only a single incident angle in the 'YZ' plane" + 
                            " (k_y!=0). Please use multiple simulations instead.")
        if len(np.unique(all_inc_angles)) == 1 and self.ky0 is not None:
            warnings.warn("'ky' is defined in Dyad but also an oblique incidence is used. "+
                          "The oblique incidence will be overridden, "+
                          "leading to potentially incorrect results.")
        
        if np.abs(struct.geometry.T[1].max()) != 0:
            raise Exception("2D structure meshpoints must be in the plane Y=0.")
        
        if struct.normalization != 1:
            raise Exception("2D supports so far only cubic mesh.")
            
        if len(struct.geometry) > 0:
            z_min = struct.geometry.T[2].min()
            z_max = struct.geometry.T[2].max()
            ## if interface 1/2 exists, check if entire structure is below or above
            if self.n1_material.__name__ != self.n2_material.__name__:
                if z_min<0 and z_max>0:
                    warnings.warn("Structure in-between substrate and middle-layer. " +
                                  "This is not supported and will most likely falsify the simulation.")
            
            ## if interface 2/3 exists, check if entire structure is below or above
            if self.n2_material.__name__ != self.n3_material.__name__:
                if z_min<self.spacing and z_max>self.spacing:
                    warnings.warn("Structure in-between middle and top cladding layer. " +
                                  "This is not supported and will most likely falsify the simulation.")
            
        return True
    
    
    def getConfigDictG(self, wavelength, struct, efield):
        ## all data need to be same dtype, must be cast to correct type inside numba functions
        from pyGDM2 import fields
        conf_dict = numba.typed.Dict.empty(key_type=numba.types.unicode_type,
                                           value_type=numba.types.complex64)
        
        conf_dict['eps1'] = np.complex64(self.n1_material.epsilon(wavelength))
        conf_dict['eps2'] = np.complex64(self.n2_material.epsilon(wavelength))
        conf_dict['eps3'] = np.complex64(self.n3_material.epsilon(wavelength))
        conf_dict['spacing'] = np.complex64(self.spacing)
        
        if self.ky0 is not None:
            conf_dict['ky0'] = np.complex64(self.ky0)
        elif efield.field_generator.__name__ == fields.plane_wave.__name__:
            if 'inc_plane' in efield.kwargs_permutations[0].keys():
                if efield.kwargs_permutations[0]['inc_plane'].lower() in ['yz', 'zy']:
                    
                    ## k-vector component parallel to y
                    inc_angle = np.pi * efield.kwargs_permutations[0]['inc_angle'] / 180.0
                    n2 = (self.n2_material.epsilon(wavelength)**0.5)
                    k_y = np.sin(inc_angle) * 2*np.pi * np.sqrt(n2**2) / wavelength
                    conf_dict['ky0'] = np.complex64(k_y)
                else:
                    conf_dict['ky0'] = np.complex64(0) # ky=0 --> inc. normal to XY plane
            else:
                conf_dict['ky0'] = np.complex64(0) # ky=0 --> inc. normal to XY plane
        elif efield.field_generator.__name__ == fields.planewave.__name__:
            conf_dict['ky0'] = np.complex64(0) # ky=0 --> inc. normal to XY plane
        elif efield.field_generator.__name__ == field_dipole_line_2d.__name__:
            if 'ky0' in efield.kwargs_permutations[0].keys():
                conf_dict['ky0'] = np.complex64(efield.kwargs_permutations[0]['ky0'])
            else:
                conf_dict['ky0'] = np.complex64(0)
        else:
            raise ValueError("2D-incompatible incident field. Set `ky0` to override this error.")
        
        ## return a numba typed dictionary of "complex64" type,
        ## can be used to pass configuration to the green's functions
        return conf_dict
    
    
    def getEnvironmentIndices(self, wavelength, geo):
        """set environment layers refractive index at 'wavelength'"""
        self.n1 = self.n1_material.epsilon(wavelength)**0.5
        self.n2 = self.n2_material.epsilon(wavelength)**0.5
        self.n3 = self.n3_material.epsilon(wavelength)**0.5
        
        ## environment epsilon at every meshpoint
        eps_env = np.zeros(len(geo), dtype=self.dtypec)
        eps_env[geo.T[2] > self.spacing] = self.n3_material.epsilon(wavelength)
        eps_env[np.logical_and(0 <= geo.T[2], geo.T[2] <= self.spacing)] = self.n2_material.epsilon(wavelength)
        eps_env[geo.T[2] < 0] = self.n1_material.epsilon(wavelength)
        
        return eps_env
    
    
    def getSelfTermEE(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        
        k0 = 2.0*np.pi / float(wavelength)
        k02 = k0**2
        
        if struct.normalization == 0:
            norm_xz = 0
            norm_y = 0
        else:
            from scipy.special import hankel1
            ky0 = 0    # no y-component of wavevector supported so far
            kr = np.sqrt(eps_env*k02 - ky0**2)
            h11 = hankel1(1, kr*struct.step/np.sqrt(np.pi))
            norm01 = (struct.step/np.sqrt(np.pi) * h11/kr + 2j / (np.pi*kr**2))
            
            norm_xz_nonrad = -1 * struct.normalization / (2.0 * struct.step**2 * eps_env)
            norm_y_nonrad = 0
            norm_xz_rad = 1j*np.pi * (2*k02 - kr**2/eps_env) * \
                               norm01 / (4*struct.step**2)
            norm_y_rad =  1j*np.pi * (k02 - ky0**2/eps_env) * \
                               norm01 / (2*struct.step**2)
                           
            norm_xz = 4.0 * np.pi * (norm_xz_nonrad + norm_xz_rad)
            norm_y = 4.0 * np.pi * (norm_y_nonrad + norm_y_rad)
        
        
        self_term_tensors_EE = np.zeros([len(eps_env), 3, 3], dtype=struct.dtypec)
        self_term_tensors_EE[:,0,0] = norm_xz
        self_term_tensors_EE[:,1,1] = norm_y
        self_term_tensors_EE[:,2,2] = norm_xz
        
        return self_term_tensors_EE
        
    
    def getSelfTermHE(self, wavelength, struct):
        struct.setWavelength(wavelength)
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        self_term_tensors_HE = np.zeros([len(eps_env), 3, 3], dtype=self.dtypec)
        
        return self_term_tensors_HE
        
        
    def getPolarizabilityTensor(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        eps = struct.epsilon_tensor 
        
        S_cell_norm = struct.step**2 / float(struct.normalization)
        
        eps_env_tensor = np.zeros(eps.shape, dtype=self.dtypec)
        eps_env_tensor[:,0,0] = eps_env
        eps_env_tensor[:,1,1] = eps_env
        eps_env_tensor[:,2,2] = eps_env
        
        ## --- isotropic polarizability
        alphatensor = np.asfortranarray((eps - eps_env_tensor) * 
                                      S_cell_norm / (4.0 * np.pi), dtype=self.dtypec)
        
        return alphatensor


### --- dipole-line incident field
def field_dipole_line_2d(pos, env_dict, wavelength, 
                    x0,z0, mx,my,mz, ky0=0,
                    dp_type="electric", returnField='E', 
                    R_farfield_approx=9000):
    """field emitted by a 2D dipole line at (x0,z0) with complex amplitude (mx,my,mz)
    
    The dipole line is infinitely long in y-direction
    
    mandatory kwargs along with `wavelength` are: `x0`, `z0`, `mx`, `my`, `mz`
    
    To take into account a dielectric interface, a quasistatic
    mirror-charge approximation is used
    
    
    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
    
    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain either "eps_env" or ["eps1", "eps2", "eps3", "spacing""]. 
    
    wavelength : float
        Wavelength in nm
    
    x0,z0 : float
        dipole-line position
          
    mx,my,mz : float
        x/y/z amplitude of elec. dipole vector
    
    ky0 : float, default: 0
        parallel wavevector component of emitter-line
    
    dp_type : str, default: "electric"
        type of dipole emitter: "electric" or "magnetic"
    
    returnField : str, default: 'E'
        if 'E': returns electric field; if 'B' or 'H': magnetic field
          
    Returns
    -------
    E0/H0:   Complex field at each position ( (Ex,Ey,Ez)-tuples )
    
    Notes
    -----
    
    for 2D free-space propagators, see e.g. 
    Martin & Piller
    *Electromagnetic scattering in polarizable backgrounds*
    Phys. Rev. E 58, 3909–3915 (1998)
    
    """
    
    if 'eps_env' in env_dict.keys():
        eps1 = eps2 = eps3 = env_dict['eps_env']
        spacing = 5000
    else:
        eps1 = env_dict['eps1']
        eps2 = env_dict['eps2']
        eps3 = env_dict['eps3']
        spacing = np.float32(env_dict['spacing'].real)
    
    
    R1 = np.array([x0, 0, z0])  # emitter location
    p = np.array([mx, my, mz])   # emitter dipole moment
    
    Ex = np.zeros(len(pos), dtype=DTYPE_C)
    Ey = np.zeros(len(pos), dtype=DTYPE_C)
    Ez = np.zeros(len(pos), dtype=DTYPE_C)
    
    ## calc propagator
    for i, R2 in enumerate(pos):
        if np.linalg.norm(R1-R2)<0.01:
            warnings.warn("Divergence of Green's tensor in line-dipole field: Evaluation position too close to dipole line. Field is set to zero.")
            xx, yy, zz, xy, xz, yx, yz, zx, zy = [np.complex64(0)]*9
        else:
            if ((dp_type.lower()=="electric" and returnField.lower() == 'e') or 
                (dp_type.lower()=="magnetic" and returnField.lower() == 'h')):
                ## --- emitted electric field
                ## mirror-charge NF approximation
                xx, yy, zz, xy, xz, yx, yz, zx, zy \
                        = Gtot_EE_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=ky0)
            else:
                ## --- emitted magnetic field (or electric field by m-dipole)
                xx, yy, zz, xy, xz, yx, yz, zx, zy \
                        = G_HE_123_2D(R1, R2, wavelength, eps1, eps2, eps3, spacing, ky0=ky0)
        
        ## propagate the dipole
        G = np.array([[xx,xy,xz],
                      [yx,yy,yz],
                      [zx,zy,zz]])
        E = np.matmul(G, p)
        
        Ex[i] = E[0]
        Ey[i] = E[1]
        Ez[i] = E[2]
    
    return np.transpose([Ex, Ey, Ez])






# encoding: utf-8
"""
Example script, comparing pyGDM with Mie theory for a constant ref.-index 
dielectric sphere (n=2)

"""
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt
import copy


from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core
from pyGDM2 import tools
from pyGDM2 import linear
from pyGDM2 import visu


## --- load pre-calculated Mie-data
wl_mie, qext_mie, qsca_mie = np.loadtxt("scat_mie_n2_D300nm.txt").T



#==============================================================================
# pyGDM setup
#==============================================================================
## ---------- Setup scale-factors for differntly fine meshes
## --- cubic

#factors_cube = [0.8, 1.0, 1.25, 1.5, 1.75]    # those add. values used in paper
factors_cube = [0.8, 1.0, 1.25]


## --- hexagonal
#factors_hex = [0.75, 1.0, 1.2, 1.38, 1.55]   # those add. values used in paper
factors_hex = [0.75, 1.0, 1.2]




#==============================================================================
# Setup incident field
#==============================================================================
field_generator = fields.planewave
## log-interval spectrum (denser at low lambda):
wavelengths = np.exp(np.linspace(np.log(300), np.log(1000), 30))
kwargs = dict(theta = [0.0])
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)




#==============================================================================
# Setup geometry (sphere D=300nm) / environment
#==============================================================================
material = materials.dummy(2.0)
n1, n2 = 1.0, 1.0     # vacuum environment

## --- cubic mesh
sim_cube = []
for fact in factors_cube:
    step = 37.5/fact
    radius = 4.*fact     # 37.5*4 = 150nm  -->  D=300nm
    geometry = structures.sphere(step, R=radius, mesh='cube')
    
    struct = structures.struct(step, geometry, material, n1,n2, 
                                   structures.get_normalization('cube'))
    
    sim = core.simulation(struct, efield) 
    sim_cube.append( copy.deepcopy(sim) )


## --- hexagonal mesh
sim_hex = []
for fact in factors_hex:
    step = 37.5/fact
    radius = 4.*fact
    geometry = structures.sphere(step, R=radius, mesh='hex', ORIENTATION=2)
    
    struct = structures.struct(step, geometry, material, n1,n2, 
                                   structures.get_normalization('hex'))
    
    sim = core.simulation(struct, efield) 
    sim_hex.append( copy.deepcopy(sim) )






#%%
#==============================================================================
# run the simulations
#==============================================================================
## --- cubic mesh
a_ext_cube = []
labels_cube = []
for sim in sim_cube:
    plt.figure()
    plt.title('(cube) N_dipoles = {}'.format(len(sim.struct.geometry)))
    visu.structure(sim)
    print('(cube) ----- N_dipoles =', len(sim.struct.geometry), end='')
    
    E = core.scatter(sim, method='lu', verbose=False)
#    E = core.scatter(sim, method='cupy', matrix_setup='numba', verbose=False)
    
    ## extinction spectrum
    field_kwargs = tools.get_possible_field_params_spectra(sim)[0]
    wl, spec = tools.calculate_spectrum(sim, field_kwargs, linear.extinct)
    a_ext = spec.T[0]
    a_geo = tools.get_geometric_cross_section(sim)
    
    a_ext_cube.append(a_ext/a_geo)
    labels_cube.append(len(sim.struct.geometry))
    print('done.')


## --- hexagonal mesh
a_ext_hex = []
labels_hex = []
for sim in sim_hex:
    plt.figure()
    plt.title('(hex) N_dipoles = {}'.format(len(sim.struct.geometry)))
    visu.structure(sim)
    print('(hex) ----- N_dipoles =', len(sim.struct.geometry), end='')
    
    E = core.scatter(sim, method='lu', verbose=False)
#    E = core.scatter(sim, method='cupy', matrix_setup='numba', verbose=False)

    ## extinction spectrum
    field_kwargs = tools.get_possible_field_params_spectra(sim)[0]
    wl, spec = tools.calculate_spectrum(sim, field_kwargs, linear.extinct)
    a_ext = spec.T[0]
    a_geo = tools.get_geometric_cross_section(sim)
    
    a_ext_hex.append(a_ext/a_geo)
    labels_hex.append(len(sim.struct.geometry))
    print('done.')





#%%
#==============================================================================
# plot
#==============================================================================
plt.figure(figsize=(11,4))

## --- cubic
plt.subplot(121)
plt.plot(wl_mie, qext_mie, 'r--', dashes=[2,1],label='Mie')

colors = plt.cm.Blues( np.linspace(0.5, 1.0, len(a_ext_cube)) )
for i, (ae, lab) in enumerate(zip(a_ext_cube, labels_cube)):
    c = colors[i]
    lab='N={}'.format(lab)
    plt.plot(wl, ae, color=c, lw=0.5)
    plt.scatter(wl, ae, marker='x', linewidth=1.5, color=c, label=lab)


plt.legend(loc='best', fontsize=12, title='cubic mesh')
plt.xlabel("wavelength (nm)")
plt.ylabel("extinction efficiency")
plt.xlim( [wl.min(), wl.max()] )


## --- hexagonal
plt.subplot(122)
plt.plot(wl_mie, qext_mie, 'r--', dashes=[2,1],label='Mie')

colors = plt.cm.Blues( np.linspace(0.5, 1.0, len(a_ext_hex)) )
for i, (ae, lab) in enumerate(zip(a_ext_hex, labels_hex)):
    c = colors[i]
    lab='N={}'.format(lab)
    plt.plot(wl, ae, color=c, lw=0.5)
    plt.scatter(wl, ae, marker='x', linewidth=1.5, color=c, label=lab)


plt.legend(loc='best', fontsize=12, title='haxagonal mesh')
plt.xlabel("wavelength (nm)")
plt.ylabel("extinction efficiency")
plt.xlim( [wl.min(), wl.max()] )



plt.tight_layout()
plt.show()



# encoding: utf-8
#
#Copyright (C) 2017-2023, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
python implementations of the Green's Dyads, accelerated using `numba`
"""
from __future__ import print_function
from __future__ import absolute_import

import copy
import warnings

import numpy as np
import math
import cmath

import numba

from pyGDM2.propagators import propagators


### --- 3D periodic Dyads - until now homogeneous environment only, normal incidence only
@numba.njit(cache=True)
def G0_periodic_123(R1, R2, wavelength, 
                    eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                    k_x=0.0, N_u=0, N_v=0, 
                    u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    
    for u_i in numba.prange(-N_u, N_u+1):
        for v_i in numba.prange(-N_v, N_v+1):
            _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
            _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.G0_EE_123(
                    _R1, R2, wavelength, eps1, eps2, eps3, spacing)
            xx[u_i, v_i] = _xx
            yy[u_i, v_i] = _yy
            zz[u_i, v_i] = _zz
            xy[u_i, v_i] = _xy
            xz[u_i, v_i] = _xz
            yx[u_i, v_i] = _yx
            yz[u_i, v_i] = _yz
            zx[u_i, v_i] = _zx
            zy[u_i, v_i] = _zy
    
    return np.sum(xx), np.sum(yy), np.sum(zz), \
           np.sum(xy), np.sum(xz), np.sum(yx), \
           np.sum(yz), np.sum(zx), np.sum(zy)


@numba.njit(cache=True)
def Gs_periodic_123(R1, R2, wavelength, 
                    eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                    k_x=0.0, N_u=0, N_v=0, 
                    u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    
    for u_i in numba.prange(-N_u, N_u+1):
        for v_i in numba.prange(-N_v, N_v+1):
            _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
            _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.Gs_EE_123(
                    _R1, R2, wavelength, eps1, eps2, eps3, spacing)
            xx[u_i, v_i] = _xx
            yy[u_i, v_i] = _yy
            zz[u_i, v_i] = _zz
            xy[u_i, v_i] = _xy
            xz[u_i, v_i] = _xz
            yx[u_i, v_i] = _yx
            yz[u_i, v_i] = _yz
            zx[u_i, v_i] = _zx
            zy[u_i, v_i] = _zy
    
    return np.sum(xx), np.sum(yy), np.sum(zz), \
           np.sum(xy), np.sum(xz), np.sum(yx), \
           np.sum(yz), np.sum(zx), np.sum(zy)


@numba.njit(cache=True)
def Gtot_periodic_123(R1, R2, wavelength, 
                      eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                      k_x=0.0, N_u=0, N_v=0, 
                      u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    ## ----- free space term
    xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123(R1, R2, wavelength, 
                                eps1, eps2, eps3, spacing, k_x, N_u, N_v, u, v)
    
    ## ----- surface term
    xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_periodic_123(R1, R2, wavelength, 
                                eps1, eps2, eps3, spacing, k_x, N_u, N_v, u, v)

    return xx + xxs, yy + yys, zz + zzs, \
           xy + xys, xz + xzs, yx + yxs, \
           yz + yzs, zx + zxs, zy + zys


@numba.njit(cache=True)
def G_HE_periodic(R1, R2, wavelength, 
                    eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                    k_x=0.0, N_u=0, N_v=0, 
                    u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    
    for u_i in numba.prange(-N_u, N_u+1):
        for v_i in numba.prange(-N_v, N_v+1):
            _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
            ## additional phase, use x-component of wave-vector --> incidence in XZ plane!
            exp_term_k_x = np.exp(1j* k_x * (u*u_i + v*v_i)[0])
            _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.G_HE_123(
                    _R1, R2, wavelength, eps1, eps2, eps3, spacing)
            xx[u_i, v_i] = _xx * exp_term_k_x
            yy[u_i, v_i] = _yy * exp_term_k_x
            zz[u_i, v_i] = _zz * exp_term_k_x
            xy[u_i, v_i] = _xy * exp_term_k_x
            xz[u_i, v_i] = _xz * exp_term_k_x
            yx[u_i, v_i] = _yx * exp_term_k_x
            yz[u_i, v_i] = _yz * exp_term_k_x
            zx[u_i, v_i] = _zx * exp_term_k_x
            zy[u_i, v_i] = _zy * exp_term_k_x
    
    return np.sum(xx), np.sum(yy), np.sum(zz), \
           np.sum(xy), np.sum(xz), np.sum(yx), \
           np.sum(yz), np.sum(zx), np.sum(zy)


# @numba.njit(cache=True)
# def G_EE_perdiodic(R1, R2, wavelength, conf_dict):
#     """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
#     R1: dipole position
#     R2: evaluation position
#     """
#     k_x = np.float32(conf_dict['k_x'].real)
#     # eps1 = conf_dict['eps1']
#     eps2 = conf_dict['eps2']
#     # eps3 = conf_dict['eps3']
#     # spacing = np.float32(conf_dict['spacing'].real)
#     cutoff_u = np.int(conf_dict['cutoff_u'].real)
#     cutoff_v = np.int(conf_dict['cutoff_v'].real)
#     u = np.array([np.float32(conf_dict['ux'].real), 
#                   np.float32(conf_dict['uy'].real), 
#                   np.float32(conf_dict['uz'].real)])
#     v = np.array([np.float32(conf_dict['vx'].real), 
#                   np.float32(conf_dict['vy'].real), 
#                   np.float32(conf_dict['vz'].real)])
    
#     return _G0_periodic(R1, R2, wavelength, eps2, k_x, cutoff_u, cutoff_v, u, v)


# @numba.njit(cache=True)
# def G_HE_perdiodic(R1, R2, wavelength, conf_dict):
#     """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
#     R1: dipole position
#     R2: evaluation position
#     """
#     k_x = np.float32(conf_dict['k_x'].real)
#     # eps1 = conf_dict['eps1']
#     eps2 = conf_dict['eps2']
#     # eps3 = conf_dict['eps3']
#     # spacing = np.float32(conf_dict['spacing'].real)
#     cutoff_u = np.int(conf_dict['cutoff_u'].real)
#     cutoff_v = np.int(conf_dict['cutoff_v'].real)
#     u = np.array([np.float32(conf_dict['ux'].real), 
#                   np.float32(conf_dict['uy'].real), 
#                   np.float32(conf_dict['uz'].real)])
#     v = np.array([np.float32(conf_dict['vx'].real), 
#                   np.float32(conf_dict['vy'].real), 
#                   np.float32(conf_dict['vz'].real)])
    
#     return _G0_HE_periodic(R1, R2, wavelength, eps2, k_x, cutoff_u, cutoff_v, u, v)



### --- coupling matrix gen. non-retarded 
# =============================================================================
# multi-dp-propagation / coupling matrix generators non-retarded 3D
# =============================================================================
## --- multi-dipole field propagation
@numba.njit(parallel=True)
def greens_tensor_evaluation_periodic(dp_pos, r_probe, G_func, wavelength, conf_dict, M,
                                      selfterm=None, dist_div_G=None):
    k_x = np.float32(conf_dict['k_x'].real)
    
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    
    cutoff_u = np.int(conf_dict['cutoff_u'].real)
    cutoff_v = np.int(conf_dict['cutoff_v'].real)
    u = np.array([np.float32(conf_dict['ux'].real), 
                  np.float32(conf_dict['uy'].real), 
                  np.float32(conf_dict['uz'].real)])
    v = np.array([np.float32(conf_dict['vx'].real), 
                  np.float32(conf_dict['vy'].real), 
                  np.float32(conf_dict['vz'].real)])
    
    for i in numba.prange(len(dp_pos)):   # explicit parallel loop
        _pos = dp_pos[i]
        for j in range(len(r_probe)):
            _r = r_probe[j]
            xx, yy, zz, xy, xz, yx, yz, zx, zy = G_func(
                                        _pos, _r, wavelength, eps1, eps2, eps3, 
                                        spacing, k_x, cutoff_u, cutoff_v, u, v
                                        )
            ## return list of Greens tensors
            M[i,j,0,0], M[i,j,1,1], M[i,j,2,2] = xx, yy, zz
            M[i,j,1,0], M[i,j,2,0], M[i,j,0,1] = yx, zx, xy
            M[i,j,2,1], M[i,j,0,2], M[i,j,1,2] = zy, xz, yz



@numba.njit(parallel=True, cache=True)
def t_sbs_EE_123_quasistatic_periodic(geo, wavelength, selfterms, alpha, conf_dict, M):
    k_x = np.float32(conf_dict['k_x'].real)
    
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    
    cutoff_u = np.int(conf_dict['cutoff_u'].real)
    cutoff_v = np.int(conf_dict['cutoff_v'].real)
    u = np.array([np.float32(conf_dict['ux'].real), 
                  np.float32(conf_dict['uy'].real), 
                  np.float32(conf_dict['uz'].real)])
    v = np.array([np.float32(conf_dict['vx'].real), 
                  np.float32(conf_dict['vy'].real), 
                  np.float32(conf_dict['vz'].real)])
    
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]       # "observer"
        for j in numba.prange(len(geo)):
            R1 = geo[j]   # emitter
            aj = alpha[j]
            st = selfterms[j]
            ## --- vacuum dyad
            if i==j:
                ## self term
                xx, yy, zz = st[0,0], st[1,1], st[2,2]
                xy, xz, yx = st[0,1], st[0,2], st[1,0]
                yz, zx, zy = st[1,2], st[2,0], st[2,1]
            else:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123(
                                        R1, R2, wavelength, eps1, eps2, eps3, 
                                        spacing, k_x, cutoff_u, cutoff_v, u, v
                                        )
            
            ## --- 1-2-3 surface dyad (non retarded NF approximation)
            if eps1!=eps2 or eps2!=eps3:
                xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_periodic_123(
                                        R1, R2, wavelength, eps1, eps2, eps3, 
                                        spacing, k_x, cutoff_u, cutoff_v, u, v
                                        )
                ## combined dyad
                xx, yy, zz, xy, xz, yx, yz, zx, zy = xx+xxs, yy+yys, zz+zzs, \
                                                      xy+xys, xz+xzs, yx+yxs, \
                                                      yz+yzs, zx+zxs, zy+zys
            
            ## return invertible matrix:  delta_ij*1 - G[i,j] * alpha[j]
            M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
            M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
            M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
            M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
            M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
            M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
            M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
            M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
            M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])
            if i==j:
                M[3*i+0, 3*j+0] += 1
                M[3*i+1, 3*j+1] += 1
                M[3*i+2, 3*j+2] += 1


@numba.njit(parallel=True, cache=True)
def t_sbs_HE_123_quasistatic_periodic(geo, wavelength, selfterms, alpha, conf_dict, M):
    k_x = np.float32(conf_dict['k_x'].real)
    
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    
    cutoff_u = np.int(conf_dict['cutoff_u'].real)
    cutoff_v = np.int(conf_dict['cutoff_v'].real)
    u = np.array([np.float32(conf_dict['ux'].real), 
                  np.float32(conf_dict['uy'].real), 
                  np.float32(conf_dict['uz'].real)])
    v = np.array([np.float32(conf_dict['vx'].real), 
                  np.float32(conf_dict['vy'].real), 
                  np.float32(conf_dict['vz'].real)])
    
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]        # "observer"
        st = selfterms[i]
        for j in numba.prange(len(geo)):
            R1 = geo[j]    # emitter
            aj = alpha[j]
            ## --- vacuum dyad
            if i==j:
                ## self term
                xx, yy, zz = st[0,0], st[1,1], st[2,2]
                xy, xz, yx = st[0,1], st[0,2], st[1,0]
                yz, zx, zy = st[1,2], st[2,0], st[2,1]
            else:
                xx, yy, zz, xy, xz, yx, yz, zx, zy = G_HE_periodic(
                                        R1, R2, wavelength, eps1, eps2, eps3, 
                                        spacing, k_x, cutoff_u, cutoff_v, u, v
                                        )
            
            ## return: G[i,j] * alpha[j]
            ## --- magnetic-electric part
            M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
            M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
            M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
            M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
            M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
            M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
            M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
            M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
            M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])


### --- dyad class 3D non-retarded, 1D or 2D periodic
class DyadsQuasistaticPeriodic123(propagators.DyadsBaseClass):
    __name__ = "Quasistatic 3D '1-2-3' Green's tensors with periodicity"
    
    def __init__(self, n1=None, n2=None, n3=None, spacing=5000, 
                 u=np.array([0,0,0]), v=np.array([0,0,0]), 
                 cutoff_u=10, cutoff_v=10, 
                 radiative_correction=True):
        """ 
        u or v: period vectors
        set np.array([0,0,0]) for no periodicity in this dimension
        
        incident wavevector must be in XZ plane
        
        """
        super().__init__()
        
        self.u = u.astype(self.dtypef)
        self.v = v.astype(self.dtypef)
        
        self.cutoff_u = cutoff_u
        self.cutoff_v = cutoff_v
        if np.linalg.norm(u)==0: 
            self.cutoff_u = 0
        if np.linalg.norm(v)==0: 
            self.cutoff_v = 0
            
        ## Dyads
        self.G_EE = Gtot_periodic_123
        self.G_HE = G_HE_periodic
        self.G_EE_ff = Gtot_periodic_123    # using full propagator here for the moment
        
        ## evaluate propagator routine
        self.eval_G = greens_tensor_evaluation_periodic
        
        ## coupling matrix constructor routines
        self.tsbs_EE = t_sbs_EE_123_quasistatic_periodic
        self.tsbs_HE = t_sbs_HE_123_quasistatic_periodic
        
        ## environment definition
        ## set ref. index values or material class of environment layers
        from pyGDM2 import materials
        if isinstance(n1, (int, float, complex)) and not isinstance(n1, bool):
            self.n1_material = materials.dummy(n1)
        else:
            self.n1_material = n1
        
        n2 = n2 or n1     # if None, use `n1`
        if isinstance(n2, (int, float, complex)) and not isinstance(n2, bool):
            from pyGDM2 import materials
            self.n2_material = materials.dummy(n2)
        else:
            self.n2_material = n2
            
        n3 = n3 or n2     # if None, use `n2`
        if isinstance(n3, (int, float, complex)) and not isinstance(n3, bool):
            from pyGDM2 import materials
            self.n3_material = materials.dummy(n3)
        else:
            self.n3_material = n3
        
        self.spacing = spacing
        self.radiative_correction = radiative_correction
    
    
    def __repr__(self, verbose=False):
        """description about simulation environment defined by set of dyads
        """
        out_str =  ' ------ environment -------'
        out_str += '\n ' + self.__name__
        out_str += '\n '
        out_str += '\n' + '   n3 = {}  <-- top'.format(self.n3_material.__name__)
        out_str += '\n' + '   n2 = {}  <-- structure zone (height "spacing" = {}nm)'.format(
                        self.n2_material.__name__, self.spacing)
        out_str += '\n' + '   n1 = {}  <-- substrate'.format(self.n1_material.__name__)
        return out_str

    
    def getConfigDictG(self, wavelength, struct, efield):
        if 'theta_inc' in efield.kwargs_permutations[0].keys():
            inc_angle = np.pi * efield.kwargs_permutations[0]['theta_inc'] / 180.0
        elif 'inc_angle' in efield.kwargs_permutations[0].keys():
            inc_angle = np.pi * efield.kwargs_permutations[0]['inc_angle'] / 180.0
        elif 'kSign' in efield.kwargs_permutations[0].keys():
            if efield.kwargs_permutations[0]['kSign'] == 1:
                inc_angle = 0.0    # bottom incidence
            else:
                inc_angle = np.pi  # top incidence
        else:
                inc_angle = np.pi  # top incidence
        
        
        ## all data need to be same dtype, must be cast to correct type inside numba functions
        conf_dict = numba.typed.Dict.empty(key_type=numba.types.unicode_type,
                                           value_type=numba.types.complex64)
        
        conf_dict['eps1'] = np.complex64(self.n1_material.epsilon(wavelength))
        conf_dict['eps2'] = np.complex64(self.n2_material.epsilon(wavelength))
        conf_dict['eps3'] = np.complex64(self.n3_material.epsilon(wavelength))
        conf_dict['spacing'] = np.complex64(self.spacing)
        conf_dict['ux'] = np.complex64(self.u[0])
        conf_dict['uy'] = np.complex64(self.u[1])
        conf_dict['uz'] = np.complex64(self.u[2])
        conf_dict['vx'] = np.complex64(self.v[0])
        conf_dict['vy'] = np.complex64(self.v[1])
        conf_dict['vz'] = np.complex64(self.v[2])
        conf_dict['cutoff_u'] = np.complex64(self.cutoff_u)
        conf_dict['cutoff_v'] = np.complex64(self.cutoff_v)
        
        n2 = (self.n2_material.epsilon(wavelength)**0.5)
        k_x = np.sin(inc_angle) * 2*np.pi * np.sqrt(n2**2) / wavelength
        conf_dict['k_x'] = np.complex64(k_x)
        
        ## return a numba typed dictionary of "complex64" type,
        ## can be used to pass configuration to the green's functions
        return conf_dict
    
    
    def exceptionHandling(self, struct, efield):
        """Exception handling / consistency and compatibility check
        
        check if structure and incident field generator are compatible

        Parameters
        ----------
        struct : :class:`.structures.struct`
            instance of structure class
        efield : :class:`.fields.efield`
            instance of incident field class

        Returns
        -------
        bool : True if struct and field are compatible, False if they don't fit the tensors

        """
        if efield.field_generator.__name__ not in ['plane_wave', 'planewave', 'evanescent_planewave']:
            raise Exception("periodic structures only work with plane wave illumination. " +
                            "Please use 'plane_wave', 'planewave' or 'evanescent_planewave' field generator!")
        
        all_inc_angles = []
        for k in efield.kwargs_permutations:
            if 'theta_inc' in k.keys():
                all_inc_angles.append(k['theta_inc'])
        if len(np.unique(all_inc_angles)) > 1:
            raise Exception("For technical reasons periodic simulations can treat only a single incident angle. " + 
                            "Please run multiple separate simulations.")
        
        return True
    
    
    def getEnvironmentIndices(self, wavelength, geo):
        """get environment permittivity for `wavelength` at each meshpoint"""
        self.n1 = self.n1_material.epsilon(wavelength)**0.5
        self.n2 = self.n2_material.epsilon(wavelength)**0.5
        self.n3 = self.n3_material.epsilon(wavelength)**0.5
        
        ## environment epsilon at every meshpoint
        eps_env = np.zeros(len(geo), dtype=self.dtypec)
        eps_env[geo.T[2].min() > self.spacing] = self.n3_material.epsilon(wavelength)
        eps_env[0 <= geo.T[2].min() <= self.spacing] = self.n2_material.epsilon(wavelength)
        eps_env[geo.T[2].min() < 0] = self.n1_material.epsilon(wavelength)
        
        return eps_env
        
        
    def getSelfTermEE(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        
        k0 = 2.0*np.pi / float(wavelength)
        
        if struct.normalization == 0:
            cnorm = np.zeros(len(eps_env))
        else:
            norm_nonrad = -4.0 * np.pi * struct.normalization / (3.0 * struct.step**3 * eps_env)
            
            if self.radiative_correction:
                norm_rad = 1j * 2.0 * struct.normalization * (k0**3)/3.0 * np.ones(len(norm_nonrad))
                cnorm = norm_nonrad + norm_rad
            else:
                cnorm = norm_nonrad
        
        self_term_tensors_EE = np.zeros([len(norm_nonrad), 3, 3], dtype=self.dtypec)
        self_term_tensors_EE[:,0,0] = cnorm
        self_term_tensors_EE[:,1,1] = cnorm
        self_term_tensors_EE[:,2,2] = cnorm
        
        return self_term_tensors_EE
        
    
    def getSelfTermHE(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        
        self_term_tensors_HE = np.zeros([len(eps_env), 3, 3], dtype=self.dtypec)
        
        return self_term_tensors_HE
        
        
    def getPolarizabilityTensor(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        normalization = struct.normalization
        eps = struct.epsilon_tensor 
        
        vcell_norm = struct.step**3 / float(normalization)
        
        eps_env_tensor = np.zeros(eps.shape, dtype=self.dtypec)
        eps_env_tensor[:,0,0] = eps_env
        eps_env_tensor[:,1,1] = eps_env
        eps_env_tensor[:,2,2] = eps_env
        
        ## --- isotropic polarizability
        alphatensor = np.asfortranarray((eps - eps_env_tensor) * 
                                      vcell_norm / (4.0 * np.pi), dtype=self.dtypec)
        
        return alphatensor






from __future__ import print_function
from __future__ import absolute_import

import unittest

import numpy as np

from pyGDM2 import propagators




class TestDyads(unittest.TestCase):

    def setUp(self):
    
        self.dyads = propagators.DyadsBaseClass()

        


    def test_2_efield(self):
        print(" ----- Start fields.efield test -----")
        self.assertEqual(self.dyads.G_EE, (50,50),
                         'incorrect default size')
# self.assertAlmostEqual(np.sum(np.abs(K.flatten())), self.K_abs_sum[i_sim], delta=.5)
        print("\nFinished `efield` test\n\n")
        
        
    


# =============================================================================
# run tests
# =============================================================================
if __name__ == '__main__':
    
    # ut_classes = TestClasses()
    # ut_classes.test_0_simulation()
    # ut_classes.test_1_struct()
    # ut_classes.test_2_efield()
    
    # ut_core = TestDyads()
    
    # ut_core.test_0_get_sbs()
    # ut_core.test_1_get_K()
    # ut_core.test_2_scatter()
    # # ut_core.test_GPU_0_get_sbs()
    # ut_core.test_GPU_1_get_K()
    # ut_core.test_GPU_2_scatter()
    
    dyads = propagators.DyadsQuasistatic123()
    dyads.G_EE



    # ## all tests
    # unittest.main()
    
    
    
    
    
    
    


# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Collection of 3D visualization tools for pyGDM2

all 3D plotting is done using `mayavi`

"""
from __future__ import print_function
from __future__ import absolute_import

import numpy as np

from . import tools

## Mayavi Imports
from mayavi import mlab
from mayavi.sources.builtin_surface import BuiltinSurface
from mayavi.modules.surface import Surface
from mayavi.filters.transform_data import TransformData
#from tvtk.util import ctf
#from tvtk.tools import visual








#==============================================================================
# predefined mayavi plotting functions
#==============================================================================
#==============================================================================
# Substrate
#==============================================================================
def _draw_rect(X0,X1, Y0,Y1, Z0,Z1, rotate_axis=(1,0,0), rotate_angle=(0),
                opacity=1, color=(.5,.5,.5)):
    engine = mlab.get_engine()
    def rotMat3D(axis, angle, tol=1e-12):
        """Return the rotation matrix for 3D rotation by angle `angle` degrees about an
        arbitrary axis `axis`.
        """
        t = np.radians(angle)
        x, y, z = axis
        R = (np.cos(t))*np.eye(3) +\
            (1-np.cos(t))*np.matrix(((x**2,x*y,x*z),(x*y,y**2,y*z),(z*x,z*y,z**2))) + \
            np.sin(t)*np.matrix(((0,-z,y),(z,0,-x),(-y,x,0)))
        R[np.abs(R)<tol]=0.0
        return R

    # Main code
    # Add a cubic builtin source
    rect_src = BuiltinSurface()
    engine.add_source(rect_src)
    rect_src.source = 'cube'
    rect_src.data_source.center = np.array([ (X1+X0)/2.,  (Y1+Y0)/2.,  (Z1+Z0)/2.])
    rect_src.data_source.x_length=X1-X0
    rect_src.data_source.y_length=Y1-Y0
    rect_src.data_source.z_length=Z1-Z0
    #~ rect_src.data_source.capping = False
    #~ rect_src.data_source.resolution = 250


    # Add transformation filter to rotate rect about an axis
    transform_data_filter = TransformData()
    engine.add_filter(transform_data_filter, rect_src)
    Rt = np.eye(4)
    Rt[0:3,0:3] = rotMat3D(rotate_axis, rotate_angle) # in homogeneous coordinates
    Rtl = list(Rt.flatten()) # transform the rotation matrix into a list
    transform_data_filter.transform.matrix.__setstate__({'elements': Rtl})
    transform_data_filter.widget.set_transform(transform_data_filter.transform)
    transform_data_filter.filter.update()
    transform_data_filter.widget.enabled = False   # disable the rotation control further.

    # Add surface module to the rect source
    rect_surface = Surface()
    engine.add_filter(rect_surface, transform_data_filter)
    # add color property
    rect_surface.actor.property.color = color
    rect_surface.actor.property.opacity = opacity
    
    return rect_surface








########################################################################
##                      VISUALIZATION FUNCTIONS
########################################################################
def structure(struct, scale=0.75, abs_scale=False, 
              tit='', color='auto', colorlist=None, mode='cube', 
              draw_substrate=True, substrate_size=2.0, 
              substrate_color=(0.8, 0.8, 0.9), substrate_opacity=0.5,
              axis_labels=True, material_labels=False,
              show=True, **kwargs):
    """plot structure in 3d
    
    plot the structure "struct" using 3d points. Either from list of 
    coordinates, or using a simulation definition dict as input.
    
    Parameters
    ----------
      - struct:    either simulation-dictionary or list of 3d coordinate tuples
      - scale:     symbol scaling in units of stepsize (default 0.75)
      - abs_scale: enable absolute scaling, override internal scale calculation (default: False)
      - color:     Color of scatterplot. Either "auto", or mayavi2-compatible color.
      - colorlist: List of colors for different materials (optional, by default (=`None`), use matplotlib color cycle)
      - mode:      3d symbols for plotting meshpoints. see `mlab.points3d`. e.g. 'cube' or 'sphere' (default 'cube')
      - draw_substrate: Whether or not to draw a substrate (default: True)
      - substrate_size: size of substrate with respect to structure extensions (default: 2.0)
      - substrate_color: default (0.8, 0.8, 0.9)
      - substrate_opacity: default 0.5
      - axis_labels: whether to show the X/Y/Z dimensions of the nano-object (default: True)
      - material_labels: whether or not to add material names labels (default: False)
      - show:      directly show plot (default True)
      - kwargs:    are passed to `mlab.points3d`
    
    """
    from pyGDM2 import structures
    X,Y,Z = tools.get_geometry(struct)
    step = tools.get_step_from_geometry(struct)
    
    ## -- set scatter-scaling depending on structure size
    maxdist = max([max(X)-min(X), max(Y)-min(Y)])
    if not abs_scale: scale = step*scale
    
    ## -- colors for subsets with different materials
    if color == 'auto':
        if hasattr(struct, "struct") or type(struct) == structures.struct:
            if hasattr(struct, "struct"):
                struct = struct.struct
        
            if hasattr(struct.material, '__iter__'):
                materials = [s.__name__ for s in struct.material]
                if len(set(materials)) > 1:
                    material = np.array(materials)
                    different_materials = np.unique(materials)
                    indices_subsets = []
                    for struct_fraction in different_materials:
                        indices_subsets.append(np.arange(len(material))[material==struct_fraction])
                else:
                    color = (0.3,0.3,0.3)
            else:
                color = (0.3,0.3,0.3)
        else:
            color = (0.3,0.3,0.3)
    
    
    
    
    # 3D-Structure
    if show: 
        mlab.figure( bgcolor=(1.0, 1.0, 1.0), fgcolor=(0.,0.,0.) )
    
    if draw_substrate:
        X0 = min(X) - (substrate_size/2. - 0.5) * maxdist
        Y0 = min(Y) - (substrate_size/2. - 0.5) * maxdist
        X1 = max(X) + (substrate_size/2. - 0.5) * maxdist
        Y1 = max(Y) + (substrate_size/2. - 0.5) * maxdist
        Z0 = - step
        Z1 = - step/10.
        _draw_rect(X0,X1, Y0,Y1, Z0,Z1, opacity=substrate_opacity, color=substrate_color)
    
    ## draw structure
    ## --- if multi-material: set maker colors
    if color != 'auto':
        im = mlab.points3d(X,Y,Z, mode=mode, scale_factor=scale, color=color, **kwargs)
    else:
        import matplotlib.pyplot as plt
        colors = [plt.cm.colors.to_rgb('C{}'.format(i)) for i in range(1,10)] * int(2 + len(indices_subsets)/9.)
        im = []
        for i, idx in enumerate(indices_subsets):
            col = colors[i]
            im.append(mlab.points3d(X[idx],Y[idx],Z[idx], mode=mode, scale_factor=scale, color=col, **kwargs))
            if material_labels:
                ## add space for approximately same text length and size
                N_max_char = max([len(l) for l in different_materials])
                mat_label = different_materials[i] + " "*2*(N_max_char-len(different_materials[i]))
                mlab.text(0.02, 0.95-0.05*i, mat_label,
                                color=col, opacity=0.8)
    
    if axis_labels: 
        mlab.axes(xlabel='X (nm)', ylabel='Y (nm)', zlabel='Z (nm)')
    if show: 
        mlab.title(tit)
        mlab.show()
    
    return im






    
def vectorfield(NF, struct=None, scale=1.5, abs_scale=False, 
                tit='', complex_part='real', clim=[0.0, 1.0],
                axis_labels=True, show=True, **kwargs):
    """3d quiverplot of nearfield
    
    Parameters
    ----------
     - NF:       Nearfield definition. `np.array`, containing 6-tuples:
                   (X,Y,Z, Ex,Ey,Ez), the field components being complex.
     - struct:   optional structure definition (if field is supplied in 3-tuple 
                 form without coordinates). Either `simulation` object, or list
                 of coordinate (x,y,z) tuples 
     - scale:     symbol scaling in units of stepsize (default 0.75)
     - abs_scale: enable absolute scaling, override internal scale calculation (default: False)
     - complex_part: Which part of complex field to plot. 
                     Either 'real' or 'imag'. (default: 'real')
     - axis_labels: whether to show the X/Y/Z dimensions of the nano-object (default: True)
     - show:     whether to directly show the figure (default: True)
    
    All other keyword arguments are passed to mlab's `quiver3d`.
    """
    ## case: provided structure+NF separately: NF may contain field_config dict as first element
    if len(NF) == 2:
        NF = NF[1]
    
    if len(NF.T) == 6:
        X,Y,Z, UXcplx,UYcplx,UZcplx = np.transpose(NF)
        X,Y,Z = X.real,Y.real,Z.real
    elif len(NF.T) == 3 and struct is not None:
        UXcplx,UYcplx,UZcplx = np.transpose(NF)
        X,Y,Z = tools.get_geometry(struct).real
    else:
        raise ValueError("Error: Wrong number of columns in vector field. Expected (Ex,Ey,Ez)-tuples + `simulation` object or (x,y,z, Ex,Ey,Ez)-tuples.")
    
    if complex_part.lower() == "real":
        Ex,Ey,Ez = UXcplx.real, UYcplx.real, UZcplx.real
    elif complex_part.lower() == "imag":
        Ex,Ey,Ez = UXcplx.imag, UYcplx.imag, UZcplx.imag
    else:
        raise ValueError("Error: Unknown `complex_part` argument value. Must be either 'real' or 'imag'.")
    S = np.sqrt(Ex**2 + Ey**2 + Ez**2)
    
    step = tools.get_step_from_geometry(np.array([X,Y,Z]).T)
    if not abs_scale:
        scale = step*scale
    else:
        scale = scale
    
    ## 3D-plot
    if show: 
        mlab.figure( bgcolor=(0.2,0.2,0.2), fgcolor=(1.0, 1.0, 1.0) )
    
    s = mlab.quiver3d(X,Y,Z, Ex,Ey,Ez, scalars=S, scale_factor=scale, 
                      vmin=clim[0], vmax=clim[1], **kwargs)
    
    if axis_labels: 
        mlab.axes(xlabel='X (nm)', ylabel='Y (nm)', zlabel='Z (nm)')
    if show: 
        mlab.title(tit)
        mlab.show()
    
    return s


def vectorfield_by_fieldindex(sim, field_index, which_field='E', **kwargs):
    """Wrapper to :func:`vectorfield`, using simulation object and fieldindex as input
    
    Parameters
    ----------
    sim : `simulation`
        instance of :class:`.core.simulation`
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    which_field : str, default: "E"
        "E" or "H", for electric or magnetic field
        
    All other keyword arguments are passed to :func:`vectorfield`.
    """
    NF = tools.get_field_as_list_by_fieldindex(sim, field_index, which_field)
    
    return vectorfield(NF, sim, **kwargs)
    






## 3D colorplot
def vectorfield_color(NF, complex_part='real', 
                      fieldComp='I', scale=0.5, abs_scale=False, mode='sphere',
                      tit='', axis_labels=True, show=True, **kwargs):
    """plot of scalar electric field data as 3D colorplot
    
    `vectorfield_color` is using `mlab.quiver3d` to plot colored data-points in
    order to be able to fix the size of the points while varying the color-code
    
    Parameters
    ----------
     - NF:       Nearfield definition. `np.array`, containing 6-tuples:
                   (X,Y,Z, Ex,Ey,Ez), the field components being complex.
     - complex_part: Complex part to plot. Either 'real' or 'imag' 
                       (default 'real')
     - fieldComp: default='I'. Which component to use. One of ["I", "abs", "Ex", "Ey", "Ez"].
                    if "I"or "abs" is used, `complex_part` argument has no effect.
     - scale:     symbol scaling in units of stepsize (default 0.75)
     - abs_scale: enable absolute scaling, override internal scale calculation (default: False)
     - mode:     which glyph to use (default: 'sphere', might also be e.g. 'cube')
     - tit:      optional plot title (default '')
     - axis_labels: whether to show the X/Y/Z dimensions of the nano-object (default: True)
     - show:     whether to directly show the figure (default: True)
     
    other kwargs are passed to mlabs's `quiver3d`
    """
    if len(NF.T) == 6:
        X,Y,Z, Ex,Ey,Ez = NF.T
        X,Y,Z = X.real, Y.real, Z.real
    else:
        raise ValueError("Error: Field list must contain tuples of exactly 6 elements.")
    
    if fieldComp.lower() not in ['i', 'abs']:
        if complex_part.lower() == "real":
            Ex, Ey, Ez = Ex.real, Ey.real, Ez.real
        elif complex_part.lower() == "imag":
            Ex, Ey, Ez = Ex.imag, Ey.imag, Ez.imag
        else:
            raise ValueError("Error: Unknown `complex_part` argument value. Must be either 'real' or 'imag'.")
    
    
    if fieldComp.lower() in ['i', 'abs']:
        EF = np.abs(Ex)**2 + np.abs(Ey)**2 + np.abs(Ez)**2
        if fieldComp.lower() == 'abs':
            EF = np.sqrt(EF)
    elif fieldComp.lower() == 'ex':
        EF = Ex
    elif fieldComp.lower() == 'ey':
        EF = Ey
    elif fieldComp.lower() == 'ez':
        EF = Ez
    
    step = tools.get_step_from_geometry(np.array([X,Y,Z]).T)
    if not abs_scale:
        scale = step*scale
    else:
        scale = scale
    
    ## 3D-plot
    if show: 
        mlab.figure( bgcolor=(0.2,0.2,0.2), fgcolor=(1.0, 1.0, 1.0) )
    
    s = np.ones(X.shape)
    o = np.zeros(X.shape)
    pts = mlab.quiver3d(X,Y,Z, s, o, o, scalars=EF, mode=mode, scale_factor=scale, **kwargs) 
    pts.glyph.color_mode = 'color_by_scalar'
    pts.glyph.glyph_source.glyph_source.center = [0, 0, 0]
    
    
    if axis_labels: 
        mlab.axes(xlabel='X (nm)', ylabel='Y (nm)', zlabel='Z (nm)')
    if show: 
        mlab.title(tit)
        mlab.show()
    
    return pts






def vectorfield_color_by_fieldindex(sim, field_index, which_field='E', **kwargs):
    """Wrapper to :func:`vectorfield_color`, using simulation object and fieldindex as input
    
    Parameters
    ----------
    sim : `simulation`
        instance of :class:`.core.simulation`
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    which_field : str, default: "E"
        "E" or "H", for electric or magnetic field
        
    All other keyword arguments are passed to :func:`vectorfield_color`.
    """
    NF = tools.get_field_as_list_by_fieldindex(sim, field_index, which_field)
    
    return vectorfield_color(NF, **kwargs)


def scalarfield(NF, **kwargs):
    """Wrapper to :func:`vectorfield_color`, using scalar data tuples (x,y,z,S) as input
    
    Parameters
    ----------
    NF : list of 4-tuples
        list of tuples (x,y,z,S).
        Alternatively, the scalar-field can be passed as list of 2 lists 
        containing the x/y positions and scalar values, respectively.
        ([xy-values, S], with xy-values: list of 2-tuples [x,y]; S: list of 
        scalars). This format is returned e.g. by 
        :func:`.tools.calculate_rasterscan`.
        
    All other keyword arguments are passed to :func:`vectorfield_color`.
    """
    if len(NF) == 2 and np.shape(NF[0])[1] == 2 and len(np.shape(NF[1])) == 1 \
                    and len(NF[0]) == len(NF[1]):
        NF = np.concatenate([NF[0].T, [np.zeros(len(NF[1]))], [NF[1]]]).T
    elif len(NF.T) != 4:
        NF = np.array(NF).T
        if len(NF.T) != 4:
            raise ValueError("Error: Scalar field must consist of 4-tuples (x,y,z,S).")
    NF = np.array(NF)
    
    x, y, z = NF.T[0:3]
    Ex = Ey = Ez = NF.T[3]
    NF = np.transpose([x,y,z, Ex,Ey,Ez])
    
    im = vectorfield_color(NF, complex_part='real', fieldComp='Ex', **kwargs)
    return im







###----------------------------------------------------------------------
###               FARFIELD (INTENSITY)
###----------------------------------------------------------------------
#def _sphericToCarthesian(TETA, PHI, INTENS):
#    """Transfrom from spherical to carthesian coordinates
#    
#      input:
#        TETA:    List of teta angles (one entry for each intensity)
#        PHI:     List of phi angles (one entry for each intensity)
#        INTENS:  List of intensities
#    
#      output:
#        EFFX,EFFY,EFFZ:  X,Y,Z values in carthesian space
#        norm:            normalization factor for dataset
#    """
#    dat = np.transpose([TETA, PHI, INTENS])
#    EFFX=[]
#    EFFY=[]
#    EFFZ=[]
#    I=[]; I2=[]
#    Theta=[]; Theta2=[]
#    lastT=9999
#    for d in dat:
#        if d[0]!=lastT:
#            lastT=d[0]
#            EFFX.append([]); EFFY.append([]); EFFZ.append([])
#        teta      = d[0]
#        phi       = d[1]
#        Intensity = d[2]
#        
#        EFFX[-1].append(Intensity * np.sin(teta) * np.cos(phi))
#        EFFY[-1].append(Intensity * np.sin(teta) * np.sin(phi))
#        EFFZ[-1].append(Intensity * np.cos(teta))
#    EFFX=np.array(EFFX); EFFY=np.array(EFFY); EFFZ=np.array(EFFZ)
#    norm = max([np.max(EFFX),np.max(EFFY),np.max(EFFZ)])
#    return EFFX,EFFY,EFFZ,norm
#
#
#def plotFarfield3D(TETA, PHI, INTENS, TETAc=None, PHIc=None, INTENSc=None, norm=1, 
#                   title='', show=True):
#    """Plot 3D Farfield Radiation Pattern using mayavi
#    
#    Plot a radiation pattern in 3D space.
#    
#      input:
#        TETA:    List of teta angles (one entry for each intensity)
#        PHI:     List of phi angles (one entry for each intensity)
#        INTENS:  List of intensities
#        title:   (optional) title of plot
#        show:     directly show plot (default True)
#        
#      return:
#        None
#      
#    """
#    
#    if show: mlab.figure(bgcolor=(.9,.9,.9), size=(800,600))
#    
#    EX,EY,EZ,n = _sphericToCarthesian(TETA, PHI, INTENS)
#    norm = float(n)/float(norm)
#    if TETAc is not None:
#        EXwire,EYwire,EZwire,_n = _sphericToCarthesian(TETAc, PHIc, INTENSc)
#        mlab.mesh(EXwire/norm,EYwire/norm,EZwire/norm, color=(0,0,1), representation='wireframe', opacity=0.1, line_width=0.1)
#    
#    
#    mlab.mesh(EX/norm,EY/norm,EZ/norm, colormap="Blues", opacity=0.3)
#    
#    
#    if show: mlab.show()
#
#

##----------------------------------------------------------------------
##               Oszillating Field Animation
##----------------------------------------------------------------------
def animate_vectorfield(NF, Nframes=50, show=True, scale=1.5, abs_scale=False, 
                        draw_struct=False,
                        draw_substrate=True, substrate_size=2.0, 
                        clim=[0.0, 1.0],
                        figsize=(600, 400), save_anim=False, 
                        fig=None, view=[45, 45],
                        rotate_azimuth=0, rotate_elevation=0,
                        ffmpeg_args="", mov_file="pygdm_animated_field.mp4",
                        tmp_file_prefix='_img_pygdmanim_tmp',
                        delete_pngs=True,
                        save_cycles=1, t_start=0, frame_list=None,
                        **kwargs):
    """animate vector-field in 3d
    
    animate the time-harmonic vectors of an electromagnetic field
    
    Parameters
    ----------
    NF : list of 4-tuples
        list of tuples (x,y,z,S).
        Alternatively, the scalar-field can be passed as list of 2 lists 
        containing the x/y positions and scalar values, respectively.
        ([xy-values, S], with xy-values: list of 2-tuples [x,y]; S: list of 
        scalars). This format is returned e.g. by 
        :func:`.tools.calculate_rasterscan`.
        
    Nframes : 
    
    
    scale : float, default: 1.5
        vector scaling in units of stepsize
        
    abs_scale : bool, default: False
        enable absolute scaling, override internal scale calculation
    
    draw_struct : bool, default: False
        whether to draw the structure, assuming that every point in 'NF' is a geometry meshpoint
        
    draw_substrate : bool, default: True
        whether to draw a sketch of a substrate
        
    substrate_size : float, default: 2.0
        size of substrate, if drawn. Factor relative to structure extension
        
    clim : 2 element list, default: [0.0, 1.0]
        range to use from colormap
    
    figsize : 2-tuple, default: (600, 400)
        plot size in pixels (X,Y)
    
    save_anim : bool, default: False
        whether to save a video-file of animation (requires the program 'ffmpeg')
    
    fig : (optional) mlab figure, default: None 
        optional mayavi2 "mlab" figure, which can already contain other plotted elements
    
    view : tuple, default (45, 45)
        view perspective, passed to :func:`mlab.view`: (azimuth, elevation, distance, [r_x,r_y,r_z])
    
    rotate_azimuth, rotate_elevation : int, int, default: 0, 0
        optional total rotation angle of camera during animation. Camera
        will rotate from initial angle alpha_0 to alpha_0 + rotate_angle.
        Should be used together with `save_cycles`>1.
    
    ffmpeg_args : str, default: "-b:v 1.5M -c:v libx264"
        string of command line arguments passed to 'ffmpeg' (if save_anim == True)
    
    mov_file : str, default: "pygdm_animated_field.mp4"
        movie file to save animation (if save_anim == True)
    
    tmp_file_prefix : str, default: '_img_pygdmanim_tmp'
        tmp-file prefix for movie-saving. Will be deleted after encoding of the movie.
    
    delete_pngs : bool, default: True
        If True, delete temporary pngs after movie creation
    
    save_cycles : int, default: 1
        number of harmonic field cycles to do for animation until restart. 
        Useful if camera is slowly rotated during animation.
    
    t_start : int, default: 0
        time-step to start animation at (=frame number)
        
    frame_list : list, default: None
        optional list of frame indices to use for animation. Can be used to 
        animate only a part of the time-harmonic cycle.
        
    **kwargs : 
        other keyword arguments are passed to :func:`mlab.quiver3d`
    
    
    Returns
    -------
    
    None
    
    """
    
    NF = NF.T
    if len(NF) != 6:
        raise ValueError("wrong shape of Nearfield Array. Must consist of 6-tuples: [x,y,z, Ex,Ey,Ez]")
    
    x,y,z = NF[0:3].real
    maxdist = max([max(x)-min(x), max(y)-min(y)])
    
    step = tools.get_step_from_geometry(NF[0:3].real.T)
    if not abs_scale:
        scale_quiver = step*scale
    else:
        scale_quiver = scale
        
    
    ## --- phase and length of complex field
    Exi = NF[3]
    Exr = np.absolute(Exi)
    Ax  = np.angle(Exi)
    
    Eyi = NF[4]
    Eyr = np.absolute(Eyi)
    Ay  = np.angle(Eyi)
    
    Ezi = NF[5]
    Ezr = np.absolute(Ezi)
    Az  = np.angle(Ezi)
    
    scaleF = float((Exr.max()+Eyr.max()+Ezr.max()))
    Exr /= scaleF
    Eyr /= scaleF
    Ezr /= scaleF
    
    ## --- create list of timesteps
    alambda = 100.
    omega = 2*np.pi/float(alambda)
    
    framnumbers = np.linspace(t_start, alambda+t_start, Nframes)
    if frame_list is not None:
        framnumbers = framnumbers[frame_list]
    
    ims = []
    Emax = []
    for t in framnumbers:    
        Ex = (Exr * np.cos((Ax - omega*t))).real
        Ey = (Eyr * np.cos((Ay - omega*t))).real
        Ez = (Ezr * np.cos((Az - omega*t))).real
        E = np.sqrt( Ex**2 +Ey**2 + Ez**2)            
        Emax.append(E.max())
        ims.append( [x,y,z, Ex,Ey,Ez, E] )
    for i, _tmp in enumerate(Emax):
        ims[i][-1] /= max(Emax)
    

#==============================================================================
#     The actual plot
#==============================================================================
    fig_new = False
    if fig is None:
        fig_new = True
        fig = mlab.figure( size=figsize, bgcolor=(1.0, 1.0, 1.0), fgcolor=(0.,0.,0.) )
    
    if draw_substrate:
        X0 = Y0 = min([min(x), min(y)]) - (substrate_size/2. - 0.5) * maxdist
        X1 = Y1 = max([max(x), max(y)]) + (substrate_size/2. - 0.5) * maxdist
        Z0 = - step
        Z1 = - step/10.
        _draw_rect(X0,X1, Y0,Y1, Z0,Z1, opacity=0.2, color=(0.3, 0.3, 0.9))
    
    if draw_struct:
        mlab.points3d(x,y,z, mode='cube', scale_factor=step*scale*0.25, 
                      color=(0.1,0.1,0.1), opacity=0.15)
    
    ## the field quiver-plot
    D = ims[-1]
    p3d = mlab.quiver3d(D[0],D[1],D[2], D[3],D[4],D[5], scalars=D[6], 
                        scale_factor=scale_quiver, vmin=clim[0], vmax=clim[1], **kwargs)
    p3d.glyph.color_mode = 'color_by_scalar'
    
    mlab.view(*view)
    
    
#==============================================================================
#     The actual animation, plus ffmpeg mp4-save stuff
#==============================================================================
    azimuth_angle = view[0]
    elevation_angle = view[1]
    
    @mlab.animate(delay=40)
    def anim():
        j_rep = 0
        while True:
            for i, D in enumerate(ims):
                ## update data
                p3d.mlab_source.set(x=D[0],y=D[1],z=D[2], u=D[3],v=D[4],w=D[5])
                p3d.mlab_source.scalars = D[6]
                
                ## update view (if accordingly configured)
                if rotate_azimuth!=0 or rotate_elevation!=0:
                    view[0] = azimuth_angle + (i + len(ims)*j_rep)*rotate_azimuth
                    view[1] = elevation_angle + (i + len(ims)*j_rep)*rotate_elevation
                    mlab.view(*view)
                
                if save_anim:
                    mlab.savefig("{:0>6}{}.png".format(i + len(ims)*j_rep, tmp_file_prefix))
                yield
            j_rep +=1
            
            if save_anim and j_rep >= save_cycles:
                break
            
        ## --- assemble movie
        if save_anim:
            from subprocess import check_output
#            print(check_output(["cd {}".format(image_tmp_dir)], shell=True))
            print(check_output(["ffmpeg -f image2 -r 20 -i '%06d{}.png' {} {}".format(tmp_file_prefix, ffmpeg_args, mov_file)], shell=True))
            if delete_pngs:
                print(check_output(["rm *{}.png".format(tmp_file_prefix)], shell=True))
    
    anim()
    
    if show:
        mlab.show()












from __future__ import print_function
from __future__ import absolute_import

import unittest

import numpy as np

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields
from pyGDM2 import core
from pyGDM2 import linear
from pyGDM2 import tools




class TestClasses(unittest.TestCase):

    ## ---------- Setup test-simulations
    mesh1 = 'cube'
    mesh2 = 'hex'
    step = 30.0
    
    
    geometry1 = structures.sphere(step, R=5, mesh=mesh1)
    geometry1 = structures.center_struct(geometry1)
    geometry1.T[2] += 50
    n1, n2, n3 = 1.0, 1.0, 1.0
    spacing = 5000
    material1 = materials.silicon()
    struct1 = structures.struct(step, geometry1, material1, n1,n2, 
                    structures.get_normalization(mesh1), n3=n3, spacing=spacing)
    
    
    geometry2 = structures.sphere(step, R=5, mesh=mesh2)
    geometry2 = structures.center_struct(geometry2)
    geometry2.T[2] += 50
    n1, n2, n3 = 1.5, 1.2, 2.0
    spacing = 500
    material2 = materials.gold()
    struct2 = structures.struct(step, geometry2, material2, n1,n2, 
                    structures.get_normalization(mesh2), n3=n3, spacing=spacing)
    
    
    wavelengths = [700., 900.]
    field_generator = fields.planewave
    kwargs1 = dict(theta=[0.0, 90.0])
    efield1 = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs1)
    kwargs2 = dict(theta=np.linspace(0,90,20), kSign=[-1, 1])
    efield2 = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs2)
    
    
    ## ---------- simulation initialization
    simulations = []
    simulations.append(core.simulation(struct1, efield1))
    simulations.append(core.simulation(struct2, efield2))

        
# =============================================================================
#     expected results
# =============================================================================
    struct_norm = [[-0.00015514037795505151+4.821189765644286e-07j, -0.00015514037795505151+2.2684061585953231e-07j],
                   [-0.00015236224067758457+6.818191953348513e-07j, -0.00015236224067758457+2.2684061585953231e-07j]]
    struct_len_alpha = [[len(geometry1), len(geometry1)], [len(geometry2), len(geometry2)]]
    struct_alpha_first = [[29634.824 + 1j* 132.546, 26061.098 + 1j* 47.449795], [-27280.238 + 1j* 1617.3942, -51914.94 + 1j* 3036.9595]]
    
    efield_len_wavelengths = [2, 2]
    efield_len_kwargs = [1, 2]
    efield_len_kwargs_permutations = [2, 40]
# =============================================================================


    def test_0_simulation(self):
        print(" ----- Start core.simulation test -----")
        self.assertNotEqual(self.simulations[0], self.simulations[1])
        
        print("  sim.:", end='')
        for i_sim, sim in enumerate(self.simulations):
            print(" {},".format(i_sim), end='')
            self.assertTrue(sim.dtypef in [np.float32, np.float64])
            self.assertTrue(sim.dtypec in [np.complex64, np.complex128])
            
        print("\nFinished `simulation` test\n\n")


    def test_1_struct(self):
        print(" ----- Start structures.struct test -----")
        print("  sim.:", end='')
        for i_sim, sim in enumerate(self.simulations):
            print(" {},".format(i_sim), end='')
            for i_wl, wl in enumerate(sim.efield.wavelengths):
                sim.struct.setWavelength(wl)
                norm = sim.struct.getNormalization(wl)
                self.assertAlmostEqual(norm, self.struct_norm[i_sim][i_wl], delta=5E-7)
                
                alpha = sim.struct.getPolarizability(wl)
                self.assertEqual(len(alpha), self.struct_len_alpha[i_sim][i_wl])
                self.assertAlmostEqual(alpha[0], self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                self.assertAlmostEqual(alpha.mean(), self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                
                alpha_tensor = sim.struct.getPolarizabilityTensor(wl)
                self.assertEqual(alpha_tensor[0].shape, (3,3))
                self.assertEqual(len(alpha_tensor), self.struct_len_alpha[i_sim][i_wl])
                self.assertAlmostEqual(alpha_tensor[0][0,0], self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                self.assertAlmostEqual(alpha_tensor[0][1,1], self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                self.assertAlmostEqual(alpha_tensor[0][2,2], self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                self.assertAlmostEqual(np.average(alpha_tensor, axis=0)[0,0], self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                self.assertAlmostEqual(np.average(alpha_tensor, axis=0)[1,1], self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                self.assertAlmostEqual(np.average(alpha_tensor, axis=0)[2,2], self.struct_alpha_first[i_sim][i_wl], delta=0.5)
                
#                print (norm, end='')
#                print (alpha[0].real, '+ 1j*', alpha[0].imag, end=', ')
#                print (np.array2string(alpha_tensor[0], separator=', '), end='\n')
                        
        print("\nFinished `struct` test\n\n")


    def test_2_efield(self):
        print(" ----- Start fields.efield test -----")
#        print("  sim.:", end='')

        for i_sim, sim in enumerate(self.simulations):
            print(" {},".format(i_sim), end='')
            self.assertEqual(len(sim.efield.wavelengths), self.efield_len_wavelengths[i_sim])
            self.assertEqual(len(sim.efield.kwargs), self.efield_len_kwargs[i_sim])
            self.assertEqual(len(sim.efield.kwargs_permutations), self.efield_len_kwargs_permutations[i_sim])

#            print(len(sim.efield.kwargs))
#            print(len(sim.efield.kwargs_permutations))
            
        print("\nFinished `efield` test\n\n")
        
        
    


class TestCore(unittest.TestCase):
    ## ---------- Setup test-simulations
    mesh = 'cube'
    step = 30.0
    
    geometry = structures.rect_wire(step, L=10, H=4, W=4, mesh=mesh)
    geometry = structures.center_struct(geometry)
    geometry.T[2] += 50
    
    
    
    n1, n2, n3 = 1.0, 1.0, 1.0
    spacing = 5000
    material1 = materials.dummy(2.5)
    struct1 = structures.struct(step, geometry, material1, n1,n2, 
                    structures.get_normalization(mesh), n3=n3, spacing=spacing)
    
    n1, n2, n3 = 1.5, 1.2, 2.0
    spacing = 500
    material2 = materials.silicon()
    struct2 = structures.struct(step, geometry, material2, n1,n2, 
                    structures.get_normalization(mesh), n3=n3, spacing=spacing)
    
    n1, n2, n3 = 2.5, 1.5, 1.0
    spacing = 400
    material3 = materials.gold()
    struct3 = structures.struct(step, geometry, material3, n1,n2, 
                    structures.get_normalization(mesh), n3=n3, spacing=spacing)
    
    n1, n2, n3 = 1.0, 1.5, 1.2
    spacing = 1500
    material4 = materials.silver()
    struct4 = structures.struct(step, geometry, material4, n1,n2, 
                    structures.get_normalization(mesh), n3=n3, spacing=spacing)
    
    
    wavelengths = [500.]                      # one single wavelength
    field_generator1 = fields.planewave
    kwargs1 = dict(theta=[0.0, 90.0])
    efield1 = fields.efield(field_generator1, wavelengths=wavelengths, kwargs=kwargs1)
    field_generator2 = fields.gaussian
    kwargs2 = dict(theta=[0.0, 90.0], spotsize=300)
    efield2 = fields.efield(field_generator2, wavelengths=wavelengths, kwargs=kwargs2)
    
    
    ## ---------- simulation initialization
    simulations = []
    for struct in [struct1, struct2, struct3, struct4]:
        simulations.append(core.simulation(struct, efield1))
    for struct in [struct1, struct2, struct3, struct4]:
        simulations.append(core.simulation(struct, efield2))
    for sim in simulations:
        sim.struct.setWavelength(sim.efield.wavelengths[0])
        sim.struct.getNormalization(sim.efield.wavelengths[0])
        sim.struct.getPolarizability(sim.efield.wavelengths[0])
    
    
# =============================================================================
#     expected results
# =============================================================================
    SBS_abs_sum = [6114.483 , 14756.186 , 3529.033 , 6873.047]
    K_abs_sum = [2069.2148 , 5252.8887 , 3431.2114 , 64333.88 , 2069.2148 , 5252.8887 , 3431.2114 , 64333.88]
    Escatter_abs_sum = [[ 201.93906 , 129.47525 ], [ 206.45422 , 105.12331 ], [ 109.123825 , 136.35785 ], [ 243.5806 , 323.61307 ], 
                        [ 225.66444 , 138.24681 ], [ 192.05823 , 102.62782 ], [ 92.782394 , 123.09842 ], [ 184.3924 , 305.2096 ]]
# =============================================================================


# =============================================================================
# CPU tests
# =============================================================================
    def test_0_get_sbs(self):
        print(" ----- Start core.get_side_by_side test -----")
        for i_sim, sim in enumerate(self.simulations[:4]):
            print(" {},".format(i_sim), end='')
            SBS = core.get_side_by_side(sim, sim.efield.wavelengths[0])
#            print(np.sum(np.abs(SBS.flatten())), ', ', end='')
            self.assertAlmostEqual(np.sum(np.abs(SBS.flatten())), self.SBS_abs_sum[i_sim], delta=.1)
            
            SBS = core.get_SBS_numba(sim, sim.efield.wavelengths[0])
            self.assertAlmostEqual(np.sum(np.abs(SBS.flatten())), self.SBS_abs_sum[i_sim], delta=.1)
            
        print("\nFinished get_sbs test\n\n")
        
    
    def test_1_get_K(self):
        print(" ----- Start core.get_general_propagator test -----", end='')
        methods = ["scipyinv", "numpyinv"]
        matrix_setup_methods = ["fortran", "numba"]
        for i_m, method in enumerate(methods):
            print("\n  testing '{}' inversion...".format(method), end='')
            for i_ms, matrix_setup in enumerate(matrix_setup_methods):
                print("\n    - with '{}' matrix setup method...".format(matrix_setup), end='')
                for i_sim, sim in enumerate(self.simulations):
                    print(" {},".format(i_sim), end='')
                    K = core.get_general_propagator(sim, wavelength=sim.efield.wavelengths[0],
                                                    method=method, verbose=0, 
                                                    matrix_setup=matrix_setup)
#                    print(np.sum(np.abs(K.flatten())), ', ', end='')
                    self.assertAlmostEqual(np.sum(np.abs(K.flatten())), self.K_abs_sum[i_sim], delta=.5)
        print("\nFinished get_general_propagator - CPU test\n\n")


    def test_2_scatter(self):
        print(" ----- Start core.scatter test -----", end='')
        methods = ["lu", "scipyinv"]
        matrix_setup_methods = ["fortran", "numba"]
        for i_m, method in enumerate(methods):
            print("\n  testing '{}' inversion...".format(method), end='')
            for i_ms, matrix_setup in enumerate(matrix_setup_methods):
                print("\n    - with '{}' matrix setup method...".format(matrix_setup))
                for i_sim, sim in enumerate(self.simulations):
                    if i_sim==0:
                        print("         testing plane wave illumination... ", end='')
                    if i_sim==4:
                        print("\n         testing gaussian illumination... ", end='')
                    print(" {},".format(i_sim), end='')
                    E = core.scatter(sim, method=method, matrix_setup=matrix_setup, verbose=0)
#                    print("[", np.sum(np.abs(E[0][1].flatten())), ', ', end='')
#                    print(np.sum(np.abs(E[1][1].flatten())), '], ', end='')
                    E_abs_sum = [np.sum(np.abs(E[0][1].flatten())), np.sum(np.abs(E[1][1].flatten()))]
                    self.assertAlmostEqual(E_abs_sum[0], self.Escatter_abs_sum[i_sim][0], delta=.002)
                    self.assertAlmostEqual(E_abs_sum[1], self.Escatter_abs_sum[i_sim][1], delta=.002)
        print("\nFinished scatter - CPU test\n\n")



# =============================================================================
# CUDA tests
# =============================================================================
    # def test_GPU_0_get_sbs(self):
    #     print(" ----- Start core.get_side_by_side -  CUDA test -----")
    #     for i_sim, sim in enumerate(self.simulations[:4]):
    #         print(" {},".format(i_sim), end='')
    #         SBS = core.get_SBS_cuda(sim, sim.efield.wavelengths[0])
    #         self.assertAlmostEqual(np.sum(np.abs(SBS.flatten())), self.SBS_abs_sum[i_sim], delta=.1)
            
    #     print("\nFinished get_sbs CUDA test\n")


    def test_GPU_1_get_K(self):
        print(" ----- Start core.get_general_propagator -  CUDA test -----")
        for i_sim, sim in enumerate(self.simulations):
            print(" {},".format(i_sim), end='')
            K = core.get_general_propagator(sim, wavelength=sim.efield.wavelengths[0],
                                            method="cuda", 
                                            matrix_setup="fortran", verbose=0)
            K = K.get()
#            print(np.sum(np.abs(K.flatten())), ', ', end='')
            self.assertAlmostEqual(np.sum(np.abs(K.flatten())), self.K_abs_sum[i_sim], delta=.5)
                        
        print("\nFinished get_general_propagator - CUDA test\n\n")
    
    
    # def test_GPU_2_scatter(self):
    #     print(" ----- Start core.scatter - CUDA test -----")
    #     for i_sim, sim in enumerate(self.simulations):
    #         if i_sim==0:
    #             print("         testing plane wave illumination... ", end='')
    #         if i_sim==4:
    #             print("\n         testing gaussian illumination... ", end='')
    #         print(" {},".format(i_sim), end='')
    #         E = core.scatter(sim, method="cuda", matrix_setup="cuda", verbose=0)
    #         E_abs_sum = [np.sum(np.abs(E[0][1].flatten())), np.sum(np.abs(E[1][1].flatten()))]
    #         self.assertAlmostEqual(E_abs_sum[0], self.Escatter_abs_sum[i_sim][0], delta=.002)
    #         self.assertAlmostEqual(E_abs_sum[1], self.Escatter_abs_sum[i_sim][1], delta=.002)
    #     print("\nFinished scatter - GPU test\n\n")
        




# =============================================================================
# run tests
# =============================================================================
if __name__ == '__main__':
    
    # ut_classes = TestClasses()
    # ut_classes.test_0_simulation()
    # ut_classes.test_1_struct()
    # ut_classes.test_2_efield()
    
    ut_core = TestCore()
    ut_core.test_0_get_sbs()
    ut_core.test_1_get_K()
    ut_core.test_2_scatter()
    # ut_core.test_GPU_0_get_sbs()
    ut_core.test_GPU_1_get_K()
    ut_core.test_GPU_2_scatter()
    
    
    # ## all tests
    # unittest.main()
    
    
    
    
    
    
    


# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    Demonstrate how to implement a user-defined problem. 
    The example problem is the maximization of the scattering cross-section
    of a simple plasmonic nano-structure under plane wave illumination with 
    fixed wavelength and linear polarization angle.

"""
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields
from pyGDM2 import linear
from pyGDM2 import tools

from pyGDM2 import core

from pyGDM2.EO.models import CrossAntenna
from pyGDM2.EO.core import run_eo





#==============================================================================
# Implement a simple problem: Maximize 2x scattering efficiency
#==============================================================================
from pyGDM2.EO.problems import BaseProblem

class ownMultiObjectiveProblem(BaseProblem):
    """maximize scattering efficiency (Q_sc) of nano-structure at two wavelengths"""
    
    def __init__(self, model):
        """
        In the Problem constructor, the constructor of the parent 
        class "BaseProblem" must be called, passing the `model` object.:
        """
        super(self.__class__, self).__init__(model)
        
    
    def objective_function(self, params):
        """calculate the scat. efficiency for 2 incident field configs
        
        This method implements the multi-objective problem: 
        The objective function returns an array with several objective
        function values (the "objective vector").
        Just as in a single-objective scenario, `objective_function` 
        is function of the free parameters `params`.
        
        Use the `field_index` parameter to select the different incident
        field configurations. In this example problem we use the first
        and last available `field_index`.
        
        """
        ## --- generate the structure geometry as function of `params`
        self.model.generate_structure(params)
        
        ## --- GDM simulation and cross-section calculation
        core.scatter(self.model.sim, verbose=0)
        
        ## --- calculate scattering efficiency 
        ##     take first and second last field config 
        ##     (e.g. two crossed polarizations at two wavelengths)
        i_field1 = 0 
        i_field2 = len(self.model.sim.E) - 1
        
        ext_cs1, sca_cs1, abs_cs1 = linear.extinct(self.model.sim, field_index=i_field1)
        ext_cs2, sca_cs2, abs_cs2 = linear.extinct(self.model.sim, field_index=i_field2)
        geom_cs = tools.get_geometric_cross_section(self.model.sim)
        
        ## --- scattering efficieny (scat eff. = scat CS / geometric CS)
        return [ float(sca_cs1/geom_cs), float(sca_cs2/geom_cs) ]





if __name__ == '__main__':
    #==============================================================================
    # Setup pyGDM part
    #==============================================================================
    ## ---------- Setup structure
    mesh = 'cube'
    step = 20
    material = materials.gold()  # structure material
    n1, n2 = 1.0, 1.0            # in vacuum
    
    ## --- Empty dummy-geometry, will be replaced on run-time by EO trial geometries
    geometry = []       
    struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))
    
    
    ## ---------- Setup incident field
    field_generator = fields.planewave        # planwave excitation
    kwargs = dict(theta = [0, 90])            # 2 polarizations
    wavelengths = [600, 1400]                 # 2 wavelengths
    efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)
    
    
    ## ---------- Simulation initialization
    sim = core.simulation(struct, efield)
    
    
    #==============================================================================
    # setup evolutionary optimization
    #==============================================================================
    ## --- output folder and file-names
    results_folder = 'eo_out'
    results_suffix = 'test_MO_problem'
    
    ## --- structure model and optimization problem
    limits_W   = [2, 20]
    limits_L   = [2, 20] 
    limits_pos = []
    height = 3
    model = CrossAntenna(sim, limits_W, limits_L, limits_pos, height)
    
    problem = ownMultiObjectiveProblem(model)
    
    
    ## --- filename to save results 
    results_filename = 'emo_test.eo'
    
    ## --- size of population
    population = 100          # Nr of individuals
    
    ## --- stop criteria
    max_time = 600            # seconds
    max_iter = 50            # max. iterations
    max_nonsuccess = 10      # max. consecutive iterations without improvement
    
    ## --- other config
    generations = 1          # generations to evolve between status reports
    plot_interval = 1
    save_all_generations = True
    
    ##  Use multi-objective algorithm "moead" (decomposition into single obj. DE)
    import pygmo as pg
    algorithm = pg.moead
    algorithm_kwargs = dict(weight_generation='grid', decomposition='bi')  # optional kwargs
    algorithm = pg.maco
    algorithm_kwargs = dict()  # optional kwargs
#    algorithm_kwargs = dict(weight_generation='low discrepancy', decomposition='weighted')  # optional kwargs
#    algorithm_kwargs = dict(neighbours=20, preserve_diversity=False)  # optional kwargs
    
    ## --- "NSGA2" EMO algorithm
#    algorithm = pg.nsga2
#    algorithm_kwargs = dict(m=0.1)  # optional kwarg: 10% mutation chance
    
    
    
    eo_dict = run_eo(problem,
                     population=population,
                     algorithm=algorithm,
                     plot_interval=plot_interval, 
                     generations=generations, 
                     max_time=max_time, max_iter=max_iter, max_nonsuccess=max_nonsuccess,
                     filename=results_filename)
    
    
    
print("\n\nOptimization finished. Results were saved to 'emo_test.eo'. " + \
      "For the analysis and plotting of the results, please run the follow-up example script: " + \
      "'example_eo_03b_multi_objective_analyze.py'.")
    
    
    
    
    



# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
    pyGDM2 - a full-field electrodynamical solver python toolkit.
    Based on the Green Dyadic Method.
    
    Contributors:
        - P.R. Wiecha (maintainer)
        - Ch. Girard
        - A. Arbouet
        - C. Majorel
        - A. Patoux
        - G. Colas des Francs
        - Y. Brûlé
        - S. Garrigou
    
"""
__name__ = 'pyGDM2'
__version__ = '1.1.6'
__date__ = "05/17/2024"   # MM/DD/YYY
__author__ = 'Peter R. Wiecha'

__all__ = ["core", "materials", "structures", "fields", 
           "propagators", "propagators_2D", "propagators_periodic",
           "linear", "nonlinear", 
           "tools", "visu", 
           "EO"]


# make modules available at package level
# from pyGDM2 import core
# from pyGDM2 import materials
# from pyGDM2 import structures
# from pyGDM2 import fields
# from pyGDM2 import propagators
# from pyGDM2 import linear
# from pyGDM2 import multipole
# from pyGDM2 import nonlinear
# from pyGDM2 import electron
# from pyGDM2 import tools
# from pyGDM2 import visu



## compatibility with deprecated transition modules:
from pyGDM2 import core as core_py
from pyGDM2 import fields as fields_py
from pyGDM2 import linear as linear_py




# encoding: utf-8
"""
Collection of problems for the EO submodule of pyGDM2

    Copyright (C) 2017-2024, P. R. Wiecha

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    

"""
from __future__ import print_function
from __future__ import absolute_import

import numpy as np

import warnings


from .. import core
from .. import linear
from .. import tools




class BaseProblem(object):
    """Base pyGDM2 problem
    
    Problem classes must call the `BaseProblem` constructor with at least 
    an instance of a model-class (see :class:`.models.BaseModel`)
    
    Parameters
    ----------
    model : instance of class implementing :class:`.models.BaseModel`
            Definition of structure model and pyGDM-simulation
    
    field_index : int, default: 0
        field_index to use from pyGDM simulation.
    
    maximize : bool, default: True
        whether to maximize (True) or minimize (False) the fitness function
    
    """
    __name__ = 'BaseProblem'
    
    def __init__(self, model=None, field_index=0, maximize=True):
        
        if model is not None:
            ## --- main simulation definitions
            self.model = model
            self.sim = model.sim
            self.field_index = field_index
            self.maximize = maximize
            
            ## -- problem boundaries
            lbnds, ubnds = self.model.get_bounds()
            self.set_bounds(lbnds, ubnds)
        else:
            raise ValueError("No valid model provided. Please init the Problem with a valid geometry model instance.")
        
        
    
#==============================================================================
# Mandatory reinmplementations for specific problems:
#==============================================================================
    def objective_function(self, params):
        """To be reimplemented! Evaluates the objective function"""
        raise NotImplementedError("'problems.BaseProblem.objective_function' not re-implemented!")


#==============================================================================
# Optionally reinmplementations
#==============================================================================
    def equality_constraints(self, params):
        """Optional (nonlinear) equality constraints (as 1D array)
        
        Equality constraints are regarded satisfied if == 0
        """
        return []
        
    def inequality_constraints(self, params):
        """Optional (nonlinear) inequality constraints (as 1D array)
        
        Inequality constraints are regarded satisfied if <= 0
        """
        return []
        
    def get_name(self):
        """Return problem object name"""
        return self.__name__

    def get_extra_info(self):
        """Return extra info, e.g. a more detailed problem description"""
        return ""

    def get_nobj(self):
        """number of objectives"""
        return len(self.fitness(self.lbnds))
    
#==============================================================================
# Internally used
#==============================================================================    
    def set_bounds(self, lbnds, ubnds):
        self.lbnds = lbnds
        self.ubnds = ubnds
    
    def get_bounds(self):
        return (self.lbnds, self.ubnds)

    def fitness(self, dv):
        """Virtual objective function, single objective (PyGMO requirement)"""
        
        ## --- target function evaluation
        fitness = self.objective_function(dv)
        if not hasattr(fitness, '__iter__'):
            fitness = [fitness, ]
        if self.maximize:
            fitness = [-1*f for f in fitness]
        
        ## --- constraint functions
        equality_constraints = self.equality_constraints(dv)  # ==0 --> satisfied
        if not hasattr(equality_constraints, '__iter__'):
            equality_constraints = [equality_constraints, ]
        
        inequality_constraints = self.inequality_constraints(dv)  # <=0 --> satisfied
        if not hasattr(inequality_constraints, '__iter__'):
            inequality_constraints = [inequality_constraints, ]
        
        return_vector = np.concatenate( [ fitness, equality_constraints, inequality_constraints ] )
        
        return return_vector

    
        
    
    
    
    
    

# =============================================================================
# Pre-defined Problems
# =============================================================================
class ProblemScat(BaseProblem):
    """optimize for scat./ext./abs. cross-section or efficiency
    
    Parameters
    ----------
    model : instance of class implementing :class:`.models.BaseModel`
        Definition of structure model and pyGDM-simulation
    
    field_index : int, default: 0
        field_index to use from pyGDM simulation.
    
    opt_target : str, default: 'Qscat'
        Optimization target. 
        One of ['Qscat', 'Qext', 'Qabs', 'CSscat', 'CSext', 'CSabs']
         - CS: cross-section (absolute value)
         - Q: efficiency (CS divided by geometrical CS)
    
    maximize : bool, default: True
        whether to maximize (True) or minimize (False) the fitness function
    
    """
    
    def __init__(self, model, field_index=0, opt_target='Qscat', 
                 maximize=True):
        """constructor"""
        super(self.__class__, self).__init__(model, field_index, 
                                         maximize=maximize)
        self.opt_target = opt_target
    
    
    def objective_function(self, params):
        """evaluate directionality ratio
        """
        self.model.generate_structure(params)
        
        ## --- GDM simulation and cross-section calculation
        core.scatter(self.model.sim, verbose=0)
        ext_cs, sca_cs, abs_cs = linear.extinct(self.model.sim, field_index=self.field_index)
        
        ## --- scattering
        if self.opt_target == "Qscat":
            geo_sect = tools.get_geometric_cross_section(self.model.sim)
            val = float(sca_cs) / geo_sect
        elif self.opt_target == "CSscat":
            val = float(sca_cs)
        
        ## --- extinction
        elif self.opt_target == "Qext":
            geo_sect = tools.get_geometric_cross_section(self.model.sim)
            val = float(ext_cs) / geo_sect
        elif self.opt_target == "CSext":
            val = float(ext_cs)
        
        ## --- absorption
        elif self.opt_target == "Qabs":
            geo_sect = tools.get_geometric_cross_section(self.model.sim)
            val = float(abs_cs) / geo_sect
        elif self.opt_target == "CSabs":
            val = float(abs_cs)
        
        return val


    def get_extra_info(self):
        return "\n\tMaximization of scattering cross section"






class ProblemNearfield(BaseProblem):
    """optimize near-field enhancement ($|E|^2$ or $|B|^2$)

    opt_target defines wether the E-field ('E') or the B-field ('B') 
    shall be maximized (or optionally minimized)
    
    Parameters
    ----------
    model : instance of class implementing :class:`.models.BaseModel`
        Definition of structure model and pyGDM-simulation
    
    field_index : int, default: 0
        field_index to use from pyGDM simulation.
    
    r_probe : 3-tuple, default: (0,0,0)
        defines the (x,y,z) position where field enhancemnt is to be
        maximized / minimized
    
    opt_target : str, default: 'E'
        Optimization target. Electric or magnetic field intensity. 
        One of ['E', 'B'].
    
    maximize : bool, default: True
        whether to maximize (True) or minimize (False) the fitness function
    
    """
    
    def __init__(self, model, field_index=0,
                 r_probe=(0,0,0), opt_target='E',
                 maximize=True):
        """constructor"""
        super(self.__class__, self).__init__(model, field_index, 
                                         maximize=maximize)
        
        self.r_probe = np.transpose( [r_probe] )
        self.opt_target = opt_target.lower()
        
        if self.opt_target not in ['e', 'b', 'h']:
            raise ValueError("'opt_target' must be one of ['e', 'b', 'h'].")
    
    
    def objective_function(self, params):
        """evaluate field intensity"""
        self.model.generate_structure(params)
        
        ## --- GDM simulation and cross-section calculation
        core.scatter(self.model.sim, verbose=0)
        Es, Etot, Bs, Btot = linear.nearfield(self.model.sim, 
                                              field_index=self.field_index,
                                              r_probe=self.r_probe)
        
        if self.opt_target.lower() == "e":
            a = Es
        elif self.opt_target.lower() in ["b", "h"]:
            a = Bs
        
        I_NF = np.abs(a[0][3]**2 + a[0][4]**2 + a[0][5]**2)
        
        return I_NF


    def get_extra_info(self):
        return "\n\tMaximization of near-field intensity"






        
class ProblemDirectivity(BaseProblem):
    """Problem to optimize directionality of scattering from nanostructure
    
    Use `EO.tools.calculate_solid_angle_by_dir_index` to define 'dir_index' 
    for the desired target solid angle
    
    
    Parameters
    ----------
    
    model : instance of class implementing :class:`.models.BaseModel`
        Definition of structure model and pyGDM-simulation
    
    field_index : list of int, default: 0
        "field_index" in case of multiple pyGDM-simulation configurations
    
    dir_index : list of int; list of list of int, default: [5]
        which solid-angle elements to consider for optimization.
        If list of lists of int (e.g. [[4], [5], [1,2]]), will run 
        multi-objective optimization using each of the sub-lists as target
        solid angle
    
    which_field : str, default: 'e_sc'
        optimize using one of ['e_sc', 'e_tot', 'e0']. 
         - 'e_sc': scattered field
         - 'e_tot': total field (= e0 + e_sc)
         - 'e0': fundamental field (rather for testing) 
    
    kwargs_farfield : dict, default: dict(Nteta=3, Nphi=5, tetamin=0, tetamax=np.pi/2.)
        kwargs, passed to :func:`pyGDM2.linear.farfield`, defining the farfield 
        scattering calculation. Required arguments are: Nteta, Nphi, tetamin, tetamax
    
    consider_dS (False), averaging (True), absoluteI (False)
        Additional flags (defaults in paranthesis): 
             - consider_dS: correct (True) solid angle integration or simple sum (False)
             - averaging: avgerage (True) or integrate (False) solid angle intensity
             - absoluteI: maximize total intensity through selected solid angle instead of ratio
    
    kwargs_scatter : dict, default: {}
        additional kwargs, passed to :func:`pyGDM2.core.sactter`
    
    maximize : bool, default: True
        whether to maximize (True) or minimize (False) the fitness function
    
    """
    def __init__(self, model, field_index=0,
                       dir_index=[5], which_field='e_sc',
                       kwargs_farfield=dict(Nteta=3, Nphi=5, tetamin=0, tetamax=np.pi/2.),
                       consider_dS=False, averaging=True, absoluteI=False, 
                       kwargs_scatter={}, maximize=True):
        """constructor"""
        ## --- directionality problem setup
        if type(dir_index) not in (list, tuple, np.ndarray):
            self.dir_index = [dir_index]
        else:
            self.dir_index = dir_index
        
        if type(field_index) not in (list, tuple, np.ndarray):
            field_index = [field_index]
        else:
            field_index = field_index
        
        ## --- number of objectives
        if self.list_depth(self.dir_index) == 2:
            N_obj = len(self.dir_index)
            if len(field_index) == 1:
                field_index = [field_index[0]]*N_obj
            if len(field_index) != N_obj:
                raise ValueError("Number of 'field_index' entries must exactly match number of objectives!")
        elif self.list_depth(self.dir_index) == 1:
            N_obj = 1
            field_index = [field_index]
            self.dir_index = [self.dir_index]
        else:
            raise ValueError("Wrong shape of `dirindex` input parameter. Must be of 'depth' 1 or 2.")
        
        ## --- init base class
        super(self.__class__, self).__init__(model, field_index, 
                                         maximize=maximize)
        
        self.which_field = which_field 
        
        self.consider_dS = consider_dS
        self.averaging = averaging
        
        self.absoluteI = absoluteI
        
        self.kwargs_scatter = kwargs_scatter
        self.kwargs_farfield = kwargs_farfield
        
        self.dteta = ((kwargs_farfield['tetamax']-kwargs_farfield['tetamin']) / 
                      float(kwargs_farfield['Nteta']-1))
        self.dphi  = 2.*np.pi/float(kwargs_farfield['Nphi']-1)


    def list_depth(self, L):
        """'depth' of a list"""
        depth = lambda L: isinstance(L, list) and max(map(depth, L))+1
        return depth(L)
        
        
    
    def objective_function(self, params):
        """evaluate directionality ratio"""
        self.model.generate_structure(params)
        
        ## --- main GDM simulation
        core.scatter(self.model.sim, verbose=0, **self.kwargs_scatter)
        
        ## --- iterate objective functions
        Iratios = []
        for i_obj, dir_idx in enumerate(self.dir_index):
            ## --- linear scattering to farfield, incoherent sum of all field_indices
            I_sc, I_tot, I0 = 0, 0, 0
            for di in self.field_index[i_obj]:
                tetalist, philist, _I_sc, _I_tot, _I0 = linear.farfield(
                                             self.model.sim, di, 
                                             **self.kwargs_farfield)
                I_sc += _I_sc
                I_tot += _I_tot
                I0 += _I0
            
            if self.which_field.lower() == 'e0':
                I = I0
            elif self.which_field.lower() == 'e_sc':
                I = I_sc
            elif self.which_field.lower() == 'e_tot':
                I = I_tot
            else:
                raise ValueError("'which_field' must be one of ['e0', 'e_sc', 'e_tot']!")
            
            tetalist = tetalist.flatten()
            philist = philist.flatten()
            I = I.flatten()
            
            
            ## --- Processing (weighting, ratio-calc.)
            if self.consider_dS:
                ## --- weight intensities by solid angle of surface element
                dS = self.dteta*self.dphi*np.sin(tetalist)
                dS = dS + dS.max()*0.1  # slight cheating: increase all weights by 10% of overall max to not neglect 180degree backscattering 
            else:
                dS = np.ones(tetalist.shape)
            
            
            ## --- compute either absolute intensity or directionality ratio
            if self.absoluteI:
                if self.averaging:
                    Iratio = np.average(I[dir_idx]*dS[dir_idx])
                else:
                    Iratio = np.sum(I[dir_idx]*dS[dir_idx])
            else:
                if self.averaging:
                    non_idx = np.delete(range(len(I)), dir_idx)
                    Iratio = (np.average(I[dir_idx]*dS[dir_idx]) / 
                                             np.average(I[non_idx]*dS[non_idx]) )
                else:
                    Iratio = (np.sum( (I[dir_idx]*dS[dir_idx])) / 
                                 (np.sum(I*dS) - np.sum(I[dir_idx]*dS[dir_idx])) )
            
            Iratios.append(Iratio)
            
        return Iratios


    ## --- Problem description
    def get_extra_info(self):
        return "\n\tOptimize directionality of scattering."+\
               " Number of objectives: {}".format(len(self.dir_index))











# encoding: utf-8
"""
Example script, demonstrating heat generation in a planar plasmonic object

"""
from __future__ import print_function, division

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core
from pyGDM2 import linear
from pyGDM2 import visu
from pyGDM2 import tools

import numpy as np
import matplotlib.pyplot as plt




#==============================================================================
# setup simulation
#==============================================================================

## --- Setup structure (prism h=12nm, side=120nm in water on glass substrate)
mesh = 'hex'
step = 6.
## note: for the example in the doc. paper, step = 3.0 was used.
## for sake of simulation time we will use a far coarser discretization, which
## nevertheless will give a very good approximation
geometry = structures.prism(step, NSIDE=int(120/step), H=12/step, mesh=mesh)
geometry = structures.center_struct(geometry)

material = materials.gold()
n1, n2 = 1.45, 1.33          # water on glass

struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh),
                           with_radiation_correction=1)



## ---------- Setup incident field
field_generator = fields.planewave        # planwave excitation
kwargs = dict(theta = 0.0, kSign=-1)
wavelengths = np.linspace(500, 1000, 51)
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)


## ---------- Simulation initialization
sim = core.simulation(struct, efield)

visu.structure(sim, scale=0.5, borders=10)
print("N dipoles:", len(sim.struct.geometry))



#%%
#==============================================================================
# run the simulation
#==============================================================================
E = core.scatter(sim)
#E = core.scatter(sim, method='cupy', matrix_setup='numba', verbose=True)



#%%
## spectrum
field_kwargs = tools.get_possible_field_params_spectra(sim)[0]
wl, heat_spectrum = tools.calculate_spectrum(sim, field_kwargs, linear.heat)


## heat distribution in struct at plasmon resonance (around 880nm for the coarse mesh)
idx_struct = tools.get_closest_field_index(sim, dict(wavelength=880))
print("structure-plot wavelength: {}".format(wl[idx_struct]))
q = linear.heat(sim, idx_struct, return_value='structure')
## divide by cell volume --> nW/nm^3
q.T[3] /= (sim.struct.step**3/sim.struct.normalization)

#%%
#==============================================================================
# plot
#==============================================================================
plt.figure( figsize=(8,4) )
plt.subplot2grid( (1,4), (0,0), colspan=3)

## --- spectrum
plt.plot(wl, heat_spectrum/1.0E3, color='r')
plt.xlabel("wavelength (nm)")
plt.ylabel("Q (microwatt)", color='r')


## --- heat distribution inside prism
plt.subplot(144, aspect='equal')
plt.title("lambda={}nm".format(wl[idx_struct]))
im = visu.scalarfield(q, cmap='hot', show=0, zorder=1)
plt.colorbar(im, label='heat (nW/nm^3)', orientation='horizontal', ticks=[0,0.5,1,1.5,2])
# im.set_clim([0, 2])

visu.structure_contour(sim, lw=2, color='k', dashes=[2,2],
                       input_mesh='hex1', zorder=0, borders=10, show=0)



plt.show()

